"""Unit tests of MQConnectionManager in the DIRAC.Resources.MessageQueue.MConnectionManager
   Also, test of internal functions for mq connection storage.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# ignore use of __functions, _functions
# pylint: disable=no-member, protected-access

import unittest
import mock
from DIRAC import S_OK
from DIRAC.Resources.MessageQueue.MQConnectionManager import MQConnectionManager


class TestMQConnectionManager(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None  # To show full difference between structures in  case of error
        dest = {}
        dest.update({"/queue/test1": ["producer4", "consumer1", "consumer2", "consumer4"]})
        dest.update({"/queue/test2": ["producer2", "consumer1", "consumer2"]})
        dest.update({"/topic/test1": ["producer1"]})
        dest4 = {"/queue/test3": ["producer1", "consumer2", "consumer3", "consumer4"]}
        conn1 = {"MQConnector": "TestConnector1", "destinations": dest}
        conn2 = {"MQConnector": "TestConnector2", "destinations": dest4}
        storage = {"mardirac3.in2p3.fr": conn1, "testdir.blabla.ch": conn2}
        self.mgr = MQConnectionManager(connectionStorage=storage)

    def tearDown(self):
        pass


class TestMQConnectionStorageFunctions_connectionExists(TestMQConnectionManager):
    def test_success(self):
        self.assertTrue(self.mgr._MQConnectionManager__connectionExists("mardirac3.in2p3.fr"))

    def test_failure(self):
        self.assertFalse(self.mgr._MQConnectionManager__connectionExists("nonexisting"))


class TestMQConnectionStorageFunctions_destinationExists(TestMQConnectionManager):
    def test_success(self):
        self.assertTrue(self.mgr._MQConnectionManager__destinationExists("mardirac3.in2p3.fr", "/queue/test1"))

    def test_failure(self):
        self.assertFalse(self.mgr._MQConnectionManager__destinationExists("nonexisting", "/queue/test1"))

    def test_failure2(self):
        self.assertFalse(self.mgr._MQConnectionManager__destinationExists("mardirac3.in2p3.fr", "/queue/nonexisting"))


class TestMQConnectionStorageFunctions_messengerExists(TestMQConnectionManager):
    def test_success(self):
        self.assertTrue(
            self.mgr._MQConnectionManager__messengerExists("mardirac3.in2p3.fr", "/queue/test1", "consumer2")
        )
        self.assertTrue(
            self.mgr._MQConnectionManager__messengerExists("mardirac3.in2p3.fr", "/queue/test1", "producer4")
        )

    def test_failure(self):
        self.assertFalse(self.mgr._MQConnectionManager__messengerExists("noexisting", "/queue/test1", "producer4"))

    def test_failure2(self):
        self.assertFalse(
            self.mgr._MQConnectionManager__messengerExists("mardirac3.in2p3.fr", "/queue/nonexisting", "producer4")
        )

    def test_failure3(self):
        self.assertFalse(
            self.mgr._MQConnectionManager__messengerExists("mardirac3.in2p3.fr", "/queue/test1", "producer10")
        )


class TestMQConnectionStorageFunctions_getConnection(TestMQConnectionManager):
    def test_success(self):
        expectedConn = {
            "MQConnector": "TestConnector2",
            "destinations": {"/queue/test3": ["producer1", "consumer2", "consumer3", "consumer4"]},
        }
        self.assertEqual(self.mgr._MQConnectionManager__getConnection("testdir.blabla.ch"), expectedConn)

    def test_failure(self):
        self.assertEqual(self.mgr._MQConnectionManager__getConnection("nonexisiting"), {})


class TestMQConnectionStorageFunctions_getAllConnections(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = ["testdir.blabla.ch", "mardirac3.in2p3.fr"]
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllConnections()), sorted(expectedOutput))


class TestMQConnectionStorageFunctions_getConnector(TestMQConnectionManager):
    def test_success(self):
        self.assertEqual(self.mgr._MQConnectionManager__getConnector("testdir.blabla.ch"), "TestConnector2")

    def test_failure(self):
        self.assertIsNone(self.mgr._MQConnectionManager__getConnector("nonexisiting"))


class TestMQConnectionStorageFunctions_setConnector(TestMQConnectionManager):
    def test_success(self):
        self.assertTrue(self.mgr._MQConnectionManager__setConnector("testdir.blabla.ch", "TestConnector5"))
        self.assertEqual(self.mgr._MQConnectionManager__getConnector("testdir.blabla.ch"), "TestConnector5")

    def test_failure(self):
        self.assertFalse(self.mgr._MQConnectionManager__setConnector("nonexisiting", "TestConnector3"))


class TestMQConnectionStorageFunctions_getDestinations(TestMQConnectionManager):
    def test_success(self):
        expectedDests = {
            "/queue/test1": ["producer4", "consumer1", "consumer2", "consumer4"],
            "/queue/test2": ["producer2", "consumer1", "consumer2"],
            "/topic/test1": ["producer1"],
        }
        self.assertEqual(self.mgr._MQConnectionManager__getDestinations("mardirac3.in2p3.fr"), expectedDests)

    def test_failure(self):
        self.assertEqual(self.mgr._MQConnectionManager__getDestinations("nonexisiting"), {})


class TestMQConnectionStorageFunctions_getMessengersId(TestMQConnectionManager):
    def test_success(self):
        expectedMess = ["producer4", "consumer1", "consumer2", "consumer4"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersId("mardirac3.in2p3.fr", "/queue/test1"), expectedMess
        )

    def test_success2(self):
        expectedMess2 = ["producer2", "consumer1", "consumer2"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersId("mardirac3.in2p3.fr", "/queue/test2"), expectedMess2
        )

    def test_failure(self):
        self.assertEqual(self.mgr._MQConnectionManager__getMessengersId("nonexisiting", "/queue/test2"), [])

    def test_failure2(self):
        self.assertEqual(self.mgr._MQConnectionManager__getMessengersId("mardirac3.in2p3.fr", "nonexisiting"), [])


class TestMQConnectionStorageFunctions_getMessengersIdWithType(TestMQConnectionManager):
    def test_success(self):
        expectedMess = ["producer4"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("mardirac3.in2p3.fr", "/queue/test1", "producer"),
            expectedMess,
        )

    def test_success2(self):
        expectedMess2 = ["producer2"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("mardirac3.in2p3.fr", "/queue/test2", "producer"),
            expectedMess2,
        )

    def test_success3(self):
        expectedMess = ["consumer1", "consumer2", "consumer4"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("mardirac3.in2p3.fr", "/queue/test1", "consumer"),
            expectedMess,
        )

    def test_success4(self):
        expectedMess2 = ["consumer1", "consumer2"]
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("mardirac3.in2p3.fr", "/queue/test2", "consumer"),
            expectedMess2,
        )

    def test_failure(self):
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("nonexisiting", "/queue/test2", "producer"), []
        )

    def test_failure2(self):
        self.assertEqual(
            self.mgr._MQConnectionManager__getMessengersIdWithType("mardirac3.in2p3.fr", "nonexisiting", "producer"), []
        )


class TestMQConnectionStorageFunctions_getAllMessengersInfo(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))


class TestMQConnectionStorageFunctions_getAllMessengersId(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = [
            "producer4",
            "consumer1",
            "consumer2",
            "consumer4",
            "producer2",
            "consumer1",
            "consumer2",
            "producer1",
            "producer1",
            "consumer2",
            "consumer3",
            "consumer4",
        ]
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersId()), sorted(expectedOutput))


class TestMQConnectionStorageFunctions_getAllMessengersIdWithType(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = [
            "consumer1",
            "consumer2",
            "consumer4",
            "consumer1",
            "consumer2",
            "consumer2",
            "consumer3",
            "consumer4",
        ]
        self.assertEqual(
            sorted(self.mgr._MQConnectionManager__getAllMessengersIdWithType("consumer")), sorted(expectedOutput)
        )
        expectedOutput = ["producer4", "producer2", "producer1", "producer1"]
        self.assertEqual(
            sorted(self.mgr._MQConnectionManager__getAllMessengersIdWithType("producer")), sorted(expectedOutput)
        )


class TestMQConnectionStorageFunctions_addMessenger(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer1",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(self.mgr._MQConnectionManager__addMessenger("mardirac3.in2p3.fr", "/queue/test1", "producer1"))
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_success2(self):
        # new queue
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test5/producer8",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(self.mgr._MQConnectionManager__addMessenger("mardirac3.in2p3.fr", "/queue/test5", "producer8"))
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_success3(self):
        # new connection
        expectedOutput = [
            "mytest.is.the.best/queue/test10/producer24",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(
            self.mgr._MQConnectionManager__addMessenger("mytest.is.the.best", "/queue/test10", "producer24")
        )
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_success4(self):
        #  two times
        expectedOutput = [
            "mytest.is.the.best/queue/test10/producer2",
            "mytest.is.the.best/queue/test10/producer24",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(
            self.mgr._MQConnectionManager__addMessenger("mytest.is.the.best", "/queue/test10", "producer24")
        )
        self.assertTrue(self.mgr._MQConnectionManager__addMessenger("mytest.is.the.best", "/queue/test10", "producer2"))
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_failure(self):
        # messenger already exists
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertFalse(self.mgr._MQConnectionManager__addMessenger("mardirac3.in2p3.fr", "/queue/test1", "producer4"))
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))


class TestMQConnectionStorageFunctions_removeMessenger(TestMQConnectionManager):
    def test_success(self):
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("mardirac3.in2p3.fr", "/queue/test1", "producer4")
        )
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_success2(self):
        # remove whole destination /topic/test1 cause only one element
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("mardirac3.in2p3.fr", "/topic/test1", "producer1")
        )
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_success3(self):
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
        ]
        # remove whole connection
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/test3", "producer1")
        )
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/test3", "consumer2")
        )
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/test3", "consumer3")
        )
        self.assertTrue(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/test3", "consumer4")
        )
        self.assertEqual(sorted(self.mgr._MQConnectionManager__getAllMessengersInfo()), sorted(expectedOutput))

    def test_failure(self):
        # remove nonexisting messenger
        self.assertFalse(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/test3", "producer10")
        )

    def test_failure2(self):
        # remove nonexisting destination
        self.assertFalse(
            self.mgr._MQConnectionManager__removeMessenger("testdir.blabla.ch", "/queue/nonexisting", "producer1")
        )

    def test_failure3(self):
        # remove nonexisting connection
        self.assertFalse(self.mgr._MQConnectionManager__removeMessenger("nonexisting", "/queue/test103", "producer1"))


class TestMQConnectionManager_addNewmessenger(TestMQConnectionManager):
    def test_success(self):
        result = self.mgr.addNewMessenger(mqURI="mardirac3.in2p3.fr::Queues::test1", messengerType="producer")
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "producer5")
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer5",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))

    def test_success2(self):
        result = self.mgr.addNewMessenger(mqURI="mardirac3.in2p3.fr::Topics::test1", messengerType="consumer")
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "consumer5")

    def test_success3(self):
        result = self.mgr.addNewMessenger(mqURI="testdir.blabla.ch::Queues::test3", messengerType="consumer")
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "consumer5")

    def test_success4(self):
        # connection does not exist
        result = self.mgr.addNewMessenger(mqURI="noexisting.blabla.ch::Queues::test3", messengerType="consumer")
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "consumer5")
        expectedOutput = [
            "noexisting.blabla.ch/queue/test3/consumer5",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))


class TestMQConnectionManager_startConnection(TestMQConnectionManager):
    def test_success(self):
        # existing connection
        result = self.mgr.startConnection(
            mqURI="mardirac3.in2p3.fr::Queues::test1", params={}, messengerType="producer"
        )
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "producer5")
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer5",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))

    @mock.patch("DIRAC.Resources.MessageQueue.MQConnectionManager.MQConnectionManager.createConnectorAndConnect")
    def test_success2(self, mock_createConnectorAndConnect):
        # connection does not exist
        mock_createConnectorAndConnect.return_value = S_OK("MyConnector")
        result = self.mgr.startConnection(
            mqURI="noexisting.blabla.ch::Queues::test3", params={}, messengerType="consumer"
        )
        self.assertTrue(result["OK"])
        self.assertEqual(result["Value"], "consumer5")
        expectedOutput = [
            "noexisting.blabla.ch/queue/test3/consumer5",
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))
        result = self.mgr.getConnector("noexisting.blabla.ch")
        self.assertEqual(result["Value"], "MyConnector")


class TestMQConnectionManager_stopConnection(TestMQConnectionManager):
    def test_success(self):
        result = self.mgr.stopConnection(mqURI="mardirac3.in2p3.fr::Queues::test1", messengerId="producer4")
        self.assertTrue(result["OK"])
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))

    def test_success2(self):
        result = self.mgr.stopConnection(mqURI="mardirac3.in2p3.fr::Topics::test1", messengerId="producer1")
        self.assertTrue(result["OK"])
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))

    @mock.patch("DIRAC.Resources.MessageQueue.MQConnectionManager.MQConnectionManager.unsubscribe")
    @mock.patch("DIRAC.Resources.MessageQueue.MQConnectionManager.MQConnectionManager.disconnect")
    def test_success3(self, mock_disconnect, mock_unsubscribe):
        mock_disconnect.return_value = S_OK()
        mock_unsubscribe.return_value = S_OK()
        result = self.mgr.stopConnection(mqURI="testdir.blabla.ch::Queues::test3", messengerId="consumer3")
        self.assertTrue(result["OK"])
        result = self.mgr.stopConnection(mqURI="testdir.blabla.ch::Queues::test3", messengerId="producer1")
        self.assertTrue(result["OK"])
        result = self.mgr.stopConnection(mqURI="testdir.blabla.ch::Queues::test3", messengerId="consumer2")
        self.assertTrue(result["OK"])
        result = self.mgr.stopConnection(mqURI="testdir.blabla.ch::Queues::test3", messengerId="consumer4")
        self.assertTrue(result["OK"])
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))


class TestMQConnectionManager_removeAllConnections(TestMQConnectionManager):
    @mock.patch("DIRAC.Resources.MessageQueue.MQConnectionManager.MQConnectionManager.disconnect")
    def test_success(self, mock_disconnect):
        mock_disconnect.return_value = S_OK()
        result = self.mgr.removeAllConnections()
        self.assertTrue(result["OK"])
        expectedOutput = []
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))


class TestMQConnectionManager_getAllMessengers(TestMQConnectionManager):
    def test_success(self):
        result = self.mgr.getAllMessengers()
        self.assertTrue(result["OK"])
        expectedOutput = [
            "mardirac3.in2p3.fr/queue/test1/producer4",
            "mardirac3.in2p3.fr/queue/test1/consumer1",
            "mardirac3.in2p3.fr/queue/test1/consumer2",
            "mardirac3.in2p3.fr/queue/test1/consumer4",
            "mardirac3.in2p3.fr/queue/test2/producer2",
            "mardirac3.in2p3.fr/queue/test2/consumer1",
            "mardirac3.in2p3.fr/queue/test2/consumer2",
            "mardirac3.in2p3.fr/topic/test1/producer1",
            "testdir.blabla.ch/queue/test3/producer1",
            "testdir.blabla.ch/queue/test3/consumer2",
            "testdir.blabla.ch/queue/test3/consumer3",
            "testdir.blabla.ch/queue/test3/consumer4",
        ]
        result = self.mgr.getAllMessengers()
        self.assertEqual(sorted(result["Value"]), sorted(expectedOutput))


class TestMQConnectionManager_getConnector(TestMQConnectionManager):
    def test_success(self):
        result = self.mgr.getConnector("mardirac3.in2p3.fr")
        self.assertTrue(result["OK"])

    def test_failure(self):
        result = self.mgr.getConnector("nonexistent.in2p3.fr")
        self.assertEqual(result["Message"], "Failed to get the MQConnector!")


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_addNewmessenger))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_startConnection))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_stopConnection))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_removeAllConnections))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_getAllMessengers))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionManager_getConnector))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_connectionExists))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_destinationExists))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_messengerExists))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getConnection))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getAllConnections))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getConnector))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_setConnector))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getDestinations))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getMessengersId))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getMessengersIdWithType)
    )
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_addMessenger))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_removeMessenger))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getAllMessengersInfo)
    )
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getAllMessengersId))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConnectionStorageFunctions_getAllMessengersIdWithType)
    )
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



""" The CS! (Configuration Service)


The following options can be set for the Configuration Service.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN Server
  :end-before: ##END
  :dedent: 2
  :caption: Service options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC.ConfigurationSystem.private.ServiceInterface import ServiceInterface
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Security.Properties import CS_ADMINISTRATOR
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR

gServiceInterface = None
gPilotSynchronizer = None


def initializeConfigurationHandler(serviceInfo):
    global gServiceInterface
    gServiceInterface = ServiceInterface(serviceInfo["URL"])
    return S_OK()


class ConfigurationHandler(RequestHandler):
    """The CS handler"""

    types_getVersion = []

    @classmethod
    def export_getVersion(cls):
        return S_OK(gServiceInterface.getVersion())

    types_getCompressedData = []

    @classmethod
    def export_getCompressedData(cls):
        sData = gServiceInterface.getCompressedConfigurationData()
        return S_OK(sData)

    types_getCompressedDataIfNewer = [six.string_types]

    @classmethod
    def export_getCompressedDataIfNewer(cls, sClientVersion):
        sVersion = gServiceInterface.getVersion()
        retDict = {"newestVersion": sVersion}
        if sClientVersion < sVersion:
            retDict["data"] = gServiceInterface.getCompressedConfigurationData()
        return S_OK(retDict)

    types_publishSlaveServer = [six.string_types]

    @classmethod
    def export_publishSlaveServer(cls, sURL):
        gServiceInterface.publishSlaveServer(sURL)
        return S_OK()

    types_commitNewData = [six.string_types]

    def export_commitNewData(self, sData):
        global gPilotSynchronizer
        credDict = self.getRemoteCredentials()
        if "DN" not in credDict or "username" not in credDict:
            return S_ERROR("You must be authenticated!")
        return gServiceInterface.updateConfiguration(sData, credDict["username"])

    types_forceGlobalConfigurationUpdate = []
    auth_forceGlobalConfigurationUpdate = [CS_ADMINISTRATOR]

    def export_forceGlobalConfigurationUpdate(self):
        """
        Attempt to request all the configured services to update their configuration

        :return: S_OK
        """
        return gServiceInterface.forceGlobalUpdate()

    types_writeEnabled = []

    @classmethod
    def export_writeEnabled(cls):
        return S_OK(gServiceInterface.isMaster())

    types_getCommitHistory = []

    @classmethod
    def export_getCommitHistory(cls, limit=100):
        if limit > 100:
            limit = 100
        history = gServiceInterface.getCommitHistory()
        if limit:
            history = history[:limit]
        return S_OK(history)

    types_getVersionContents = [list]

    @classmethod
    def export_getVersionContents(cls, versionList):
        contentsList = []
        for version in versionList:
            retVal = gServiceInterface.getVersionContents(version)
            if retVal["OK"]:
                contentsList.append(retVal["Value"])
            else:
                return S_ERROR("Can't get contents for version %s: %s" % (version, retVal["Message"]))
        return S_OK(contentsList)

    types_rollbackToVersion = [six.string_types]

    def export_rollbackToVersion(self, version):
        retVal = gServiceInterface.getVersionContents(version)
        if not retVal["OK"]:
            return S_ERROR("Can't get contents for version %s: %s" % (version, retVal["Message"]))
        credDict = self.getRemoteCredentials()
        if "DN" not in credDict or "username" not in credDict:
            return S_ERROR("You must be authenticated!")
        return gServiceInterface.updateConfiguration(retVal["Value"], credDict["username"], updateVersionOption=True)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import hashlib
import io
import os
import tarfile
import tempfile
import threading

import six
from six import StringIO

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.FrameworkSystem.Client.Logger import gLogger

try:
    # Python 2: "file" is built-in
    file_types = file, io.IOBase
except NameError:
    # Python 3: "file" fully replaced with IOBase
    file_types = (io.IOBase,)

gLogger = gLogger.getSubLogger("FileTransmissionHelper")


class FileHelper(object):

    __validDirections = ("toClient", "fromClient", "receive", "send")
    __directionsMapping = {"toClient": "send", "fromClient": "receive"}

    def __init__(self, oTransport=None, checkSum=True):
        self.oTransport = oTransport
        self.__checkMD5 = checkSum
        self.__oMD5 = hashlib.md5()
        self.bFinishedTransmission = False
        self.bReceivedEOF = False
        self.direction = False
        self.packetSize = 1048576
        self.__fileBytes = 0
        self.__log = gLogger.getSubLogger("FileHelper")

    def disableCheckSum(self):
        self.__checkMD5 = False

    def enableCheckSum(self):
        self.__checkMD5 = True

    def setTransport(self, oTransport):
        self.oTransport = oTransport

    def setDirection(self, direction):
        if direction in FileHelper.__validDirections:
            if direction in FileHelper.__directionsMapping:
                self.direction = FileHelper.__directionsMapping[direction]
            else:
                self.direction = direction

    def getHash(self):
        return self.__oMD5.hexdigest()

    def getTransferedBytes(self):
        return self.__fileBytes

    def sendData(self, sBuffer):
        if six.PY3 and isinstance(sBuffer, str):
            sBuffer = sBuffer.encode(errors="surrogateescape")
        if self.__checkMD5:
            self.__oMD5.update(sBuffer)
        retVal = self.oTransport.sendData(S_OK([True, sBuffer]))
        if not retVal["OK"]:
            return retVal
        retVal = self.oTransport.receiveData()
        return retVal

    def sendEOF(self):
        retVal = self.oTransport.sendData(S_OK([False, self.__oMD5.hexdigest()]))
        if not retVal["OK"]:
            return retVal
        self.__finishedTransmission()
        return S_OK()

    def sendError(self, errorMsg):
        retVal = self.oTransport.sendData(S_ERROR(errorMsg))
        if not retVal["OK"]:
            return retVal
        self.__finishedTransmission()
        return S_OK()

    def receiveData(self, maxBufferSize=0):
        retVal = self.oTransport.receiveData(maxBufferSize=maxBufferSize)
        if "AbortTransfer" in retVal and retVal["AbortTransfer"]:
            self.oTransport.sendData(S_OK())
            self.__finishedTransmission()
            self.bReceivedEOF = True
            return S_OK("")
        if not retVal["OK"]:
            return retVal
        stBuffer = retVal["Value"]
        if stBuffer[0]:
            if six.PY3 and isinstance(stBuffer[1], str):
                stBuffer[1] = stBuffer[1].encode(errors="surrogateescape")
            if self.__checkMD5:
                self.__oMD5.update(stBuffer[1])
            self.oTransport.sendData(S_OK())
        else:
            self.bReceivedEOF = True
            if self.__checkMD5 and not self.__oMD5.hexdigest() == stBuffer[1]:
                self.bErrorInMD5 = True
            self.__finishedTransmission()
            return S_OK("")
        return S_OK(stBuffer[1])

    def receivedEOF(self):
        return self.bReceivedEOF

    def markAsTransferred(self):
        if not self.bFinishedTransmission:
            if self.direction == "receive":
                self.oTransport.receiveData()
                abortTrans = S_OK()
                abortTrans["AbortTransfer"] = True
                self.oTransport.sendData(abortTrans)
            else:
                abortTrans = S_OK([False, ""])
                abortTrans["AbortTransfer"] = True
                retVal = self.oTransport.sendData(abortTrans)
                if not retVal["OK"]:
                    return retVal
                self.oTransport.receiveData()
        self.__finishedTransmission()

    def __finishedTransmission(self):
        self.bFinishedTransmission = True

    def finishedTransmission(self):
        return self.bFinishedTransmission

    def errorInTransmission(self):
        return self.bErrorInMD5

    def networkToString(self, maxFileSize=0):
        """Receive the input from a DISET client and return it as a string"""

        stringIO = StringIO()
        result = self.networkToDataSink(stringIO, maxFileSize=maxFileSize)
        if not result["OK"]:
            return result
        return S_OK(stringIO.getvalue())

    def networkToFD(self, iFD, maxFileSize=0):
        dataSink = os.fdopen(iFD, "w")
        try:
            return self.networkToDataSink(dataSink, maxFileSize=maxFileSize)
        finally:
            try:
                dataSink.close()
            except Exception as e:
                pass

    def networkToDataSink(self, dataSink, maxFileSize=0):
        if "write" not in dir(dataSink):
            return S_ERROR("%s data sink object does not have a write method" % str(dataSink))
        self.__oMD5 = hashlib.md5()
        self.bReceivedEOF = False
        self.bErrorInMD5 = False
        receivedBytes = 0
        # try:
        result = self.receiveData(maxBufferSize=maxFileSize)
        if not result["OK"]:
            return result
        strBuffer = result["Value"]
        if six.PY3 and isinstance(strBuffer, str):
            strBuffer = strBuffer.encode(errors="surrogateescape")
        receivedBytes += len(strBuffer)
        while not self.receivedEOF():
            if maxFileSize > 0 and receivedBytes > maxFileSize:
                self.sendError("Exceeded maximum file size")
                return S_ERROR("Received file exceeded maximum size of %s bytes" % (maxFileSize))
            dataSink.write(strBuffer)
            result = self.receiveData(maxBufferSize=(maxFileSize - len(strBuffer)))
            if not result["OK"]:
                return result
            strBuffer = result["Value"]
            if six.PY3 and isinstance(strBuffer, str):
                strBuffer = strBuffer.encode(errors="surrogateescape")
            receivedBytes += len(strBuffer)
        if strBuffer:
            dataSink.write(strBuffer)
        # except Exception as e:
        #   return S_ERROR("Error while receiving file, %s" % str(e))
        if self.errorInTransmission():
            return S_ERROR("Error in the file CRC")
        self.__fileBytes = receivedBytes
        return S_OK()

    def stringToNetwork(self, stringVal):
        """Send a given string to the DISET client over the network"""

        stringIO = StringIO(stringVal)

        iPacketSize = self.packetSize
        ioffset = 0
        strlen = len(stringVal)
        try:
            while (ioffset) < strlen:
                if (ioffset + iPacketSize) < strlen:
                    result = self.sendData(stringVal[ioffset : ioffset + iPacketSize])
                else:
                    result = self.sendData(stringVal[ioffset:strlen])
                if not result["OK"]:
                    return result
                if "AbortTransfer" in result and result["AbortTransfer"]:
                    self.__log.verbose("Transfer aborted")
                    return S_OK()
                ioffset += iPacketSize
            self.sendEOF()
        except Exception as e:
            return S_ERROR("Error while sending string: %s" % str(e))
        try:
            stringIO.close()
        except Exception:
            pass
        return S_OK()

    def FDToNetwork(self, iFD):
        self.__oMD5 = hashlib.md5()
        iPacketSize = self.packetSize
        self.__fileBytes = 0
        sentBytes = 0
        try:
            sBuffer = os.read(iFD, iPacketSize)
            while len(sBuffer) > 0:
                dRetVal = self.sendData(sBuffer)
                if not dRetVal["OK"]:
                    return dRetVal
                if "AbortTransfer" in dRetVal and dRetVal["AbortTransfer"]:
                    self.__log.verbose("Transfer aborted")
                    return S_OK()
                sentBytes += len(sBuffer)
                sBuffer = os.read(iFD, iPacketSize)
            self.sendEOF()
        except Exception as e:
            gLogger.exception("Error while sending file")
            return S_ERROR("Error while sending file: %s" % str(e))
        self.__fileBytes = sentBytes
        return S_OK()

    def BufferToNetwork(self, stringToSend):
        sIO = StringIO(stringToSend)
        try:
            return self.DataSourceToNetwork(sIO)
        finally:
            sIO.close()

    def DataSourceToNetwork(self, dataSource):
        if "read" not in dir(dataSource):
            return S_ERROR("%s data source object does not have a read method" % str(dataSource))
        self.__oMD5 = hashlib.md5()
        iPacketSize = self.packetSize
        self.__fileBytes = 0
        sentBytes = 0
        try:
            sBuffer = dataSource.read(iPacketSize)
            while len(sBuffer) > 0:
                dRetVal = self.sendData(sBuffer)
                if not dRetVal["OK"]:
                    return dRetVal
                if "AbortTransfer" in dRetVal and dRetVal["AbortTransfer"]:
                    self.__log.verbose("Transfer aborted")
                    return S_OK()
                sentBytes += len(sBuffer)
                sBuffer = dataSource.read(iPacketSize)
            self.sendEOF()
        except Exception as e:
            gLogger.exception("Error while sending file")
            return S_ERROR("Error while sending file: %s" % str(e))
        self.__fileBytes = sentBytes
        return S_OK()

    def getFileDescriptor(self, uFile, sFileMode):
        closeAfter = True
        if isinstance(uFile, six.string_types):
            try:
                self.oFile = open(uFile, sFileMode)
            except IOError:
                return S_ERROR("%s can't be opened" % uFile)
            iFD = self.oFile.fileno()
        elif isinstance(uFile, file_types):
            iFD = uFile.fileno()
        elif isinstance(uFile, int):
            iFD = uFile
            closeAfter = False
        else:
            return S_ERROR("%s is not a valid file." % uFile)
        result = S_OK(iFD)
        result["closeAfterUse"] = closeAfter
        return result

    def getDataSink(self, uFile):
        closeAfter = True
        if isinstance(uFile, six.string_types):
            try:
                oFile = open(uFile, "wb")
            except IOError:
                return S_ERROR("%s can't be opened" % uFile)
        elif isinstance(uFile, file_types):
            oFile = uFile
            closeAfter = False
        elif isinstance(uFile, int):
            oFile = os.fdopen(uFile, "wb")
            closeAfter = True
        elif "write" in dir(uFile):
            oFile = uFile
            closeAfter = False
        else:
            return S_ERROR("%s is not a valid file." % uFile)
        result = S_OK(oFile)
        result["closeAfterUse"] = closeAfter
        return result

    def __createTar(self, fileList, wPipe, compress, autoClose=True):
        if "write" in dir(wPipe):
            filePipe = wPipe
        else:
            filePipe = os.fdopen(wPipe, "w")
        tarMode = "w|"
        if compress:
            tarMode = "w|bz2"

        with tarfile.open(name="Pipe", mode=tarMode, fileobj=filePipe) as tar:
            for entry in fileList:
                tar.add(os.path.realpath(entry), os.path.basename(entry), recursive=True)
        if autoClose:
            try:
                filePipe.close()
            except Exception:
                pass

    def bulkToNetwork(self, fileList, compress=True, onthefly=True):
        if not onthefly:
            try:
                filePipe, filePath = tempfile.mkstemp()
            except Exception as e:
                return S_ERROR("Can't create temporary file to pregenerate the bulk: %s" % str(e))
            self.__createTar(fileList, filePipe, compress)
            try:
                fo = open(filePath, "rb")
            except Exception as e:
                return S_ERROR("Can't read pregenerated bulk: %s" % str(e))
            result = self.DataSourceToNetwork(fo)
            try:
                fo.close()
                os.unlink(filePath)
            except Exception:
                pass
            return result
        else:
            rPipe, wPipe = os.pipe()
            thrd = threading.Thread(target=self.__createTar, args=(fileList, wPipe, compress))
            thrd.start()
            response = self.FDToNetwork(rPipe)
            try:
                os.close(rPipe)
            except Exception:
                pass
            return response

    def __extractTar(self, destDir, rPipe, compress):
        filePipe = os.fdopen(rPipe, "r")
        tarMode = "r|*"
        if compress:
            tarMode = "r|bz2"
        with tarfile.open(mode=tarMode, fileobj=filePipe) as tar:
            for tarInfo in tar:
                tar.extract(tarInfo, destDir)
        try:
            filePipe.close()
        except Exception:
            pass

    def __receiveToPipe(self, wPipe, retList, maxFileSize):
        retList.append(self.networkToFD(wPipe, maxFileSize=maxFileSize))
        try:
            os.close(wPipe)
        except Exception:
            pass

    def networkToBulk(self, destDir, compress=True, maxFileSize=0):
        retList = []
        rPipe, wPipe = os.pipe()
        thrd = threading.Thread(target=self.__receiveToPipe, args=(wPipe, retList, maxFileSize))
        thrd.start()
        try:
            self.__extractTar(destDir, rPipe, compress)
        except Exception as e:
            return S_ERROR("Error while extracting bulk: %s" % e)
        thrd.join()
        return retList[0]

    def bulkListToNetwork(self, iFD, compress=True):
        filePipe = os.fdopen(iFD, "r")
        try:
            tarMode = "r|"
            if compress:
                tarMode = "r|bz2"
            entries = []
            with tarfile.open(mode=tarMode, fileobj=filePipe) as tar:
                for tarInfo in tar:
                    entries.append(tarInfo.name)
            filePipe.close()
            return S_OK(entries)
        except tarfile.ReadError as v:
            return S_ERROR("Error reading bulk: %s" % str(v))
        except tarfile.CompressionError as v:
            return S_ERROR("Error in bulk compression setting: %s" % str(v))
        except Exception as v:
            return S_ERROR("Error in listing bulk: %s" % str(v))



"""
  The Job Agent class instantiates a CE that acts as a client to a
  compute resource and also to the WMS.
  The Job Agent constructs a classAd based on the local resource description in the CS
  and the current resource status that is used for matching.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

__RCSID__ = "$Id$"

import os
import sys
import re
import time
import six

from diraccfg import CFG

from DIRAC import S_OK, S_ERROR, gConfig, rootPath, siteName
from DIRAC.Core.Utilities.ModuleFactory import ModuleFactory
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Security import Properties
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import TimeLeft
from DIRAC.Resources.Computing.ComputingElementFactory import ComputingElementFactory
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.WorkloadManagementSystem.Client.MatcherClient import MatcherClient
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Utilities.Utils import createJobWrapper
from DIRAC.WorkloadManagementSystem.Client import PilotStatus


class JobAgent(AgentModule):
    """This agent is what runs in a worker node. The pilot runs it, after having prepared its configuration."""

    def __init__(self, agentName, loadName, baseAgentName=False, properties=None):
        """Just defines some default parameters"""
        if not properties:
            properties = {}
        super(JobAgent, self).__init__(agentName, loadName, baseAgentName, properties)

        # Inner CE
        # CE type the JobAgent submits to. It can be "InProcess" or "Pool" or "Singularity".
        self.ceName = "InProcess"
        # "Inner" CE submission type (e.g. for Pool CE). It can be "InProcess" or "Singularity".
        self.innerCESubmissionType = "InProcess"
        self.computingElement = None  # The ComputingElement object, e.g. SingularityComputingElement()

        # Localsite options
        self.siteName = "Unknown"
        self.pilotReference = "Unknown"
        self.defaultProxyLength = 86400 * 5

        # Agent options
        # This is the factor to convert raw CPU to Normalized units (based on the CPU Model)
        self.cpuFactor = 0.0
        self.jobSubmissionDelay = 10
        self.fillingMode = True
        self.minimumTimeLeft = 5000
        self.stopOnApplicationFailure = True
        self.stopAfterFailedMatches = 10
        self.jobCount = 0
        self.matchFailedCount = 0
        self.extraOptions = ""

        # Timeleft
        self.initTimes = os.times()
        self.initTimeLeft = 0.0
        self.timeLeft = self.initTimeLeft
        self.timeLeftUtil = None
        self.pilotInfoReportedFlag = False

    #############################################################################
    def initialize(self):
        """Sets default parameters and creates CE instance"""

        # Disable monitoring
        self.am_disableMonitoring()

        localCE = gConfig.getValue("/LocalSite/LocalCE", self.ceName)
        if localCE != self.ceName:
            self.log.info("Defining Inner CE from local configuration", "= %s" % localCE)

        # Create backend Computing Element
        ceFactory = ComputingElementFactory()
        self.ceName = localCE.split("/")[0]  # It might be "Pool/Singularity", or simply "Pool"
        self.innerCESubmissionType = (
            localCE.split("/")[1] if len(localCE.split("/")) == 2 else self.innerCESubmissionType
        )
        ceInstance = ceFactory.getCE(self.ceName)
        if not ceInstance["OK"]:
            self.log.warn("Can't instantiate a CE", ceInstance["Message"])
            return ceInstance
        self.computingElement = ceInstance["Value"]
        self.computingElement.ceParameters["InnerCESubmissionType"] = self.innerCESubmissionType

        result = self.computingElement.getDescription()
        if not result["OK"]:
            self.log.warn("Can not get the CE description")
            return result
        if isinstance(result["Value"], list):
            ceDict = result["Value"][0]
        else:
            ceDict = result["Value"]
        self.initTimeLeft = ceDict.get("CPUTime", self.initTimeLeft)
        self.initTimeLeft = gConfig.getValue("/Resources/Computing/CEDefaults/MaxCPUTime", self.initTimeLeft)
        self.timeLeft = self.initTimeLeft

        self.initTimes = os.times()
        # Localsite options
        self.siteName = siteName()
        self.pilotReference = gConfig.getValue("/LocalSite/PilotReference", self.pilotReference)
        self.defaultProxyLength = gConfig.getValue("/Registry/DefaultProxyLifeTime", self.defaultProxyLength)
        # Agent options
        # This is the factor to convert raw CPU to Normalized units (based on the CPU Model)
        self.cpuFactor = gConfig.getValue("/LocalSite/CPUNormalizationFactor", self.cpuFactor)
        self.jobSubmissionDelay = self.am_getOption("SubmissionDelay", self.jobSubmissionDelay)
        self.fillingMode = self.am_getOption("FillingModeFlag", self.fillingMode)
        self.minimumTimeLeft = self.am_getOption("MinimumTimeLeft", self.minimumTimeLeft)
        self.stopOnApplicationFailure = self.am_getOption("StopOnApplicationFailure", self.stopOnApplicationFailure)
        self.stopAfterFailedMatches = self.am_getOption("StopAfterFailedMatches", self.stopAfterFailedMatches)
        self.extraOptions = gConfig.getValue("/AgentJobRequirements/ExtraOptions", self.extraOptions)
        # Utilities
        self.timeLeftUtil = TimeLeft()
        return S_OK()

    #############################################################################
    def execute(self):
        """The JobAgent execution method."""

        # Temporary mechanism to pass a shutdown message to the agent
        if os.path.exists("/var/lib/dirac_drain"):
            return self.__finish("Node is being drained by an operator")

        # Check if we can match jobs at all
        self.log.verbose("Job Agent execution loop")
        result = self.computingElement.available()
        if not result["OK"]:
            self.log.info("Resource is not available", result["Message"])
            return self.__finish("CE Not Available")

        ceInfoDict = result["CEInfoDict"]
        runningJobs = ceInfoDict.get("RunningJobs")
        availableSlots = result["Value"]

        if not availableSlots:
            if runningJobs:
                self.log.info("No available slots", ": %d running jobs" % runningJobs)
                return S_OK("Job Agent cycle complete with %d running jobs" % runningJobs)
            self.log.info("CE is not available (and there are no running jobs)")
            return self.__finish("CE Not Available")

        if self.jobCount:
            # Only call timeLeft utility after a job has been picked up
            self.log.info("Attempting to check CPU time left for filling mode")
            if self.fillingMode:
                self.timeLeft = self.computeCPUWorkLeft()
                self.log.info("normalized CPU units remaining in slot", self.timeLeft)
                if self.timeLeft <= self.minimumTimeLeft:
                    return self.__finish("No more time left")
                # Need to update the Configuration so that the new value is published in the next matching request
                result = self.computingElement.setCPUTimeLeft(cpuTimeLeft=self.timeLeft)
                if not result["OK"]:
                    return self.__finish(result["Message"])

                # Update local configuration to be used by submitted job wrappers
                localCfg = CFG()
                if self.extraOptions:
                    localConfigFile = os.path.join(".", self.extraOptions)
                else:
                    localConfigFile = os.path.join(rootPath, "etc", "dirac.cfg")
                localCfg.loadFromFile(localConfigFile)
                if not localCfg.isSection("/LocalSite"):
                    localCfg.createNewSection("/LocalSite")
                localCfg.setOption("/LocalSite/CPUTimeLeft", self.timeLeft)
                localCfg.writeToFile(localConfigFile)

            else:
                return self.__finish("Filling Mode is Disabled")

        # if we are here we assume that a job can be matched
        result = self.computingElement.getDescription()
        if not result["OK"]:
            return result

        # We can have several prioritized job retrieval strategies
        if isinstance(result["Value"], dict):
            ceDictList = [result["Value"]]
        elif isinstance(result["Value"], list):
            # This is the case for Pool ComputingElement, and parameter 'MultiProcessorStrategy'
            ceDictList = result["Value"]

        for ceDict in ceDictList:

            # Add pilot information
            gridCE = gConfig.getValue("LocalSite/GridCE", "Unknown")
            if gridCE != "Unknown":
                ceDict["GridCE"] = gridCE
            if "PilotReference" not in ceDict:
                ceDict["PilotReference"] = str(self.pilotReference)
            ceDict["PilotBenchmark"] = self.cpuFactor
            ceDict["PilotInfoReportedFlag"] = self.pilotInfoReportedFlag

            # Add possible job requirements
            result = gConfig.getOptionsDict("/AgentJobRequirements")
            if result["OK"]:
                requirementsDict = result["Value"]
                ceDict.update(requirementsDict)
                self.log.info("Requirements:", requirementsDict)

            self.log.verbose("CE dict", ceDict)

            # here finally calling the matcher
            start = time.time()
            jobRequest = MatcherClient().requestJob(ceDict)
            matchTime = time.time() - start
            self.log.info("MatcherTime", "= %.2f (s)" % (matchTime))
            if jobRequest["OK"]:
                break

        self.stopAfterFailedMatches = self.am_getOption("StopAfterFailedMatches", self.stopAfterFailedMatches)

        if not jobRequest["OK"]:

            # if we don't match a job, independently from the reason,
            # we wait a bit longer before trying again
            self.am_setOption("PollingTime", int(self.am_getOption("PollingTime") * 1.5))

            if re.search("No match found", jobRequest["Message"]):
                self.log.notice("Job request OK, but no match found", ": %s" % (jobRequest["Message"]))
                self.matchFailedCount += 1
                if self.matchFailedCount > self.stopAfterFailedMatches:
                    return self.__finish("Nothing to do for more than %d cycles" % self.stopAfterFailedMatches)
                return S_OK(jobRequest["Message"])
            elif jobRequest["Message"].find("seconds timeout") != -1:
                self.log.error("Timeout while requesting job", jobRequest["Message"])
                self.matchFailedCount += 1
                if self.matchFailedCount > self.stopAfterFailedMatches:
                    return self.__finish("Nothing to do for more than %d cycles" % self.stopAfterFailedMatches)
                return S_OK(jobRequest["Message"])
            elif jobRequest["Message"].find("Pilot version does not match") != -1:
                errorMsg = "Pilot version does not match the production version"
                self.log.error(errorMsg, jobRequest["Message"].replace(errorMsg, ""))
                return S_ERROR(jobRequest["Message"])
            else:
                self.log.notice("Failed to get jobs", ": %s" % (jobRequest["Message"]))
                self.matchFailedCount += 1
                if self.matchFailedCount > self.stopAfterFailedMatches:
                    return self.__finish("Nothing to do for more than %d cycles" % self.stopAfterFailedMatches)
                return S_OK(jobRequest["Message"])

        # Reset the Counter
        self.matchFailedCount = 0

        # If we are here it is because we matched a job
        matcherInfo = jobRequest["Value"]
        if not self.pilotInfoReportedFlag:
            # Check the flag after the first access to the Matcher
            self.pilotInfoReportedFlag = matcherInfo.get("PilotInfoReportedFlag", False)
        jobID = matcherInfo["JobID"]
        jobReport = JobReport(jobID, "JobAgent@%s" % self.siteName)
        matcherParams = ["JDL", "DN", "Group"]
        for param in matcherParams:
            if param not in matcherInfo:
                jobReport.setJobStatus(status=JobStatus.FAILED, minorStatus="Matcher did not return %s" % (param))
                return self.__finish("Matcher Failed")
            elif not matcherInfo[param]:
                jobReport.setJobStatus(status=JobStatus.FAILED, minorStatus="Matcher returned null %s" % (param))
                return self.__finish("Matcher Failed")
            else:
                self.log.verbose("Matcher returned", "%s = %s " % (param, matcherInfo[param]))

        jobJDL = matcherInfo["JDL"]
        jobGroup = matcherInfo["Group"]
        ownerDN = matcherInfo["DN"]

        optimizerParams = {}
        for key in matcherInfo:
            if key not in matcherParams:
                optimizerParams[key] = matcherInfo[key]

        parameters = self._getJDLParameters(jobJDL)
        if not parameters["OK"]:
            jobReport.setJobStatus(status=JobStatus.FAILED, minorStatus="Could Not Extract JDL Parameters")
            self.log.warn("Could Not Extract JDL Parameters", parameters["Message"])
            return self.__finish("JDL Problem")

        params = parameters["Value"]
        if "JobID" not in params:
            msg = "Job has not JobID defined in JDL parameters"
            jobReport.setJobStatus(status=JobStatus.FAILED, minorStatus=msg)
            self.log.warn(msg)
            return self.__finish("JDL Problem")
        else:
            jobID = params["JobID"]

        if "JobType" not in params:
            self.log.warn("Job has no JobType defined in JDL parameters")
            jobType = "Unknown"
        else:
            jobType = params["JobType"]

        if "CPUTime" not in params:
            self.log.warn("Job has no CPU requirement defined in JDL parameters")

        # Job requirements for determining the number of processors
        # the minimum number of processors requested
        processors = int(params.get("NumberOfProcessors", int(params.get("MinNumberOfProcessors", 1))))
        # the maximum number of processors allowed to the payload
        maxNumberOfProcessors = int(params.get("MaxNumberOfProcessors", 0))
        # need or not the whole node for the job
        wholeNode = "WholeNode" in params
        mpTag = "MultiProcessor" in params.get("Tags", [])

        if self.extraOptions and "dirac-jobexec" in params.get("Executable", "").strip():
            params["Arguments"] = (params.get("Arguments", "") + " " + self.extraOptions).strip()
            params["ExtraOptions"] = self.extraOptions

        self.log.verbose("Job request successful: \n", jobRequest["Value"])
        self.log.info("Received", "JobID=%s, JobType=%s, OwnerDN=%s, JobGroup=%s" % (jobID, jobType, ownerDN, jobGroup))
        self.jobCount += 1
        try:
            jobReport.setJobParameter(par_name="MatcherServiceTime", par_value=str(matchTime), sendFlag=False)

            if "BOINC_JOB_ID" in os.environ:
                # Report BOINC environment
                for thisp in ("BoincUserID", "BoincHostID", "BoincHostPlatform", "BoincHostName"):
                    jobReport.setJobParameter(
                        par_name=thisp, par_value=gConfig.getValue("/LocalSite/%s" % thisp, "Unknown"), sendFlag=False
                    )

            jobReport.setJobStatus(status=JobStatus.MATCHED, minorStatus="Job Received by Agent", sendFlag=False)
            result_setupProxy = self._setupProxy(ownerDN, jobGroup)
            if not result_setupProxy["OK"]:
                return self._rescheduleFailedJob(jobID, result_setupProxy["Message"], self.stopOnApplicationFailure)
            proxyChain = result_setupProxy.get("Value")

            # Save the job jdl for external monitoring
            self.__saveJobJDLRequest(jobID, jobJDL)

            software = self._checkInstallSoftware(jobID, params, ceDict, jobReport)
            if not software["OK"]:
                self.log.error("Failed to install software for job", "%s" % (jobID))
                errorMsg = software["Message"]
                if not errorMsg:
                    errorMsg = "Failed software installation"
                return self._rescheduleFailedJob(jobID, errorMsg, self.stopOnApplicationFailure)

            gridCE = gConfig.getValue("/LocalSite/GridCE", "")
            if gridCE:
                jobReport.setJobParameter(par_name="GridCE", par_value=gridCE, sendFlag=False)

            queue = gConfig.getValue("/LocalSite/CEQueue", "")
            if queue:
                jobReport.setJobParameter(par_name="CEQueue", par_value=queue, sendFlag=False)

            self.log.debug("Before self._submitJob() (%sCE)" % (self.ceName))
            result_submitJob = self._submitJob(
                jobID=jobID,
                jobParams=params,
                resourceParams=ceDict,
                optimizerParams=optimizerParams,
                proxyChain=proxyChain,
                jobReport=jobReport,
                processors=processors,
                wholeNode=wholeNode,
                maxNumberOfProcessors=maxNumberOfProcessors,
                mpTag=mpTag,
            )

            # Committing the JobReport before evaluating the result of job submission
            res = jobReport.commit()
            if not res["OK"]:
                resFD = jobReport.generateForwardDISET()
                if not resFD["OK"]:
                    self.log.error("Error generating ForwardDISET operation", resFD["Message"])
                elif resFD["Value"]:
                    # Here we create the Request.
                    op = resFD["Value"]
                    request = Request()
                    requestName = "jobAgent_%s" % jobID
                    request.RequestName = requestName.replace('"', "")
                    request.JobID = jobID
                    request.SourceComponent = "JobAgent_%s" % jobID
                    request.addOperation(op)
                    # This might fail, but only a message would be printed.
                    self._sendFailoverRequest(request)

            if not result_submitJob["OK"]:
                return self.__finish(result_submitJob["Message"])
            elif "PayloadFailed" in result_submitJob:
                # Do not keep running and do not overwrite the Payload error
                message = "Payload execution failed with error code %s" % result_submitJob["PayloadFailed"]
                if self.stopOnApplicationFailure:
                    return self.__finish(message, self.stopOnApplicationFailure)
                else:
                    self.log.info(message)

            self.log.debug("After %sCE submitJob()" % (self.ceName))
        except Exception as subExcept:  # pylint: disable=broad-except
            self.log.exception("Exception in submission", "", lException=subExcept, lExcInfo=True)
            return self._rescheduleFailedJob(
                jobID, "Job processing failed with exception", self.stopOnApplicationFailure
            )

        return S_OK("Job Agent cycle complete")

    #############################################################################
    def __saveJobJDLRequest(self, jobID, jobJDL):
        """Save job JDL local to JobAgent."""
        classAdJob = ClassAd(jobJDL)
        classAdJob.insertAttributeString("LocalCE", self.ceName)
        jdlFileName = jobID + ".jdl"
        jdlFile = open(jdlFileName, "w")
        jdl = classAdJob.asJDL()
        jdlFile.write(jdl)
        jdlFile.close()

    #############################################################################
    def computeCPUWorkLeft(self, processors=1):
        """
        Compute CPU Work Left in hepspec06 seconds

        :param int processors: number of processors available
        :return: cpu work left (cpu time left * cpu power of the cpus)
        """
        # Sum all times but the last one (elapsed_time) and remove times at init (is this correct?)
        cpuTime = sum(os.times()[:-1]) - sum(self.initTimes[:-1])
        result = self.timeLeftUtil.getTimeLeft(cpuTime, processors)
        if result["OK"]:
            cpuWorkLeft = result["Value"]
        else:
            self.log.warn("There were errors calculating time left using the Timeleft utility", result["Message"])
            self.log.warn("The time left will be calculated using os.times() and the info in our possession")
            cpuWorkLeft = self._getCPUWorkLeft(cpuTime)
        return cpuWorkLeft

    def _getCPUWorkLeft(self, cpuConsumed):
        """
        Compute the CPU Work Left as estimated by DIRAC using the process time
        and the CPU Power defined locally in the Local Config.

        :param int cpuConsumed: cpu time already consumed since the beginning of the execution
        :return: cpu work left (cpu time left * cpu power of the cpus)
        """
        self.log.info("Current raw CPU time consumed is %s" % cpuConsumed)
        cpuWorkleft = self.timeLeft
        if self.cpuFactor:
            cpuWorkleft = self.initTimeLeft - cpuConsumed * self.cpuFactor
        return cpuWorkleft

    #############################################################################
    def _setupProxy(self, ownerDN, ownerGroup):
        """
        Retrieve a proxy for the execution of the job
        """
        if gConfig.getValue("/DIRAC/Security/UseServerCertificate", False):
            proxyResult = self._requestProxyFromProxyManager(ownerDN, ownerGroup)
            if not proxyResult["OK"]:
                self.log.error("Failed to setup proxy", proxyResult["Message"])
                return S_ERROR("Failed to setup proxy: %s" % proxyResult["Message"])
            return S_OK(proxyResult["Value"])
        else:
            ret = getProxyInfo(disableVOMS=True)
            if not ret["OK"]:
                self.log.error("Invalid Proxy", ret["Message"])
                return S_ERROR("Invalid Proxy")

            proxyChain = ret["Value"]["chain"]
            if "groupProperties" not in ret["Value"]:
                print(ret["Value"])
                print(proxyChain.dumpAllToString())
                self.log.error("Invalid Proxy", "Group has no properties defined")
                return S_ERROR("Proxy has no group properties defined")

            groupProps = ret["Value"]["groupProperties"]
            if Properties.GENERIC_PILOT in groupProps or Properties.PILOT in groupProps:
                proxyResult = self._requestProxyFromProxyManager(ownerDN, ownerGroup)
                if not proxyResult["OK"]:
                    self.log.error("Invalid Proxy", proxyResult["Message"])
                    return S_ERROR("Failed to setup proxy: %s" % proxyResult["Message"])
                proxyChain = proxyResult["Value"]

        return S_OK(proxyChain)

    #############################################################################
    def _requestProxyFromProxyManager(self, ownerDN, ownerGroup):
        """Retrieves user proxy with correct role for job and sets up environment to
        run job locally.
        """

        self.log.info("Requesting proxy', 'for %s@%s" % (ownerDN, ownerGroup))
        token = gConfig.getValue("/Security/ProxyToken", "")
        if not token:
            self.log.verbose("No token defined. Trying to download proxy without token")
            token = False
        retVal = gProxyManager.getPayloadProxyFromDIRACGroup(ownerDN, ownerGroup, self.defaultProxyLength, token)
        if not retVal["OK"]:
            self.log.error("Could not retrieve payload proxy", retVal["Message"])
            os.system("dirac-proxy-info")
            sys.stdout.flush()
            return S_ERROR("Error retrieving proxy")

        chain = retVal["Value"]
        return S_OK(chain)

    #############################################################################
    def _checkInstallSoftware(self, jobID, jobParams, resourceParams, jobReport):
        """Checks software requirement of job and whether this is already present
        before installing software locally.
        """
        if "SoftwareDistModule" not in jobParams:
            msg = "Job has no software installation requirement"
            self.log.verbose(msg)
            return S_OK(msg)

        jobReport.setJobStatus(status=JobStatus.MATCHED, minorStatus="Installing Software", sendFlag=False)
        softwareDist = jobParams["SoftwareDistModule"]
        self.log.verbose("Found VO Software Distribution module", ": %s" % (softwareDist))
        argumentsDict = {"Job": jobParams, "CE": resourceParams}
        moduleFactory = ModuleFactory()
        moduleInstance = moduleFactory.getModule(softwareDist, argumentsDict)
        if not moduleInstance["OK"]:
            return moduleInstance

        module = moduleInstance["Value"]
        return module.execute()

    #############################################################################
    def _submitJob(
        self,
        jobID,
        jobParams,
        resourceParams,
        optimizerParams,
        proxyChain,
        jobReport,
        processors=1,
        wholeNode=False,
        maxNumberOfProcessors=0,
        mpTag=False,
    ):
        """Submit job to the Computing Element instance after creating a custom
        Job Wrapper with the available job parameters.
        """
        logLevel = self.am_getOption("DefaultLogLevel", "INFO")
        defaultWrapperLocation = self.am_getOption(
            "JobWrapperTemplate", "DIRAC/WorkloadManagementSystem/JobWrapper/JobWrapperTemplate.py"
        )

        # Add the number of requested processors to the job environment
        if "ExecutionEnvironment" in jobParams:
            if isinstance(jobParams["ExecutionEnvironment"], six.string_types):
                jobParams["ExecutionEnvironment"] = jobParams["ExecutionEnvironment"].split(";")
        jobParams.setdefault("ExecutionEnvironment", []).append("DIRAC_JOB_PROCESSORS=%d" % processors)

        jobDesc = {
            "jobID": jobID,
            "jobParams": jobParams,
            "resourceParams": resourceParams,
            "optimizerParams": optimizerParams,
            "extraOptions": self.extraOptions,
            "defaultWrapperLocation": defaultWrapperLocation,
        }
        result = createJobWrapper(log=self.log, logLevel=logLevel, **jobDesc)
        if not result["OK"]:
            return result

        wrapperFile = result["Value"]
        jobReport.setJobStatus(status=JobStatus.MATCHED, minorStatus="Submitting To CE")

        self.log.info("Submitting JobWrapper", "%s to %sCE" % (os.path.basename(wrapperFile), self.ceName))

        # Pass proxy to the CE
        proxy = proxyChain.dumpAllToString()
        if not proxy["OK"]:
            self.log.error("Invalid proxy", proxy)
            return S_ERROR("Payload Proxy Not Found")

        payloadProxy = proxy["Value"]
        submission = self.computingElement.submitJob(
            wrapperFile,
            payloadProxy,
            numberOfProcessors=processors,
            maxNumberOfProcessors=maxNumberOfProcessors,
            wholeNode=wholeNode,
            mpTag=mpTag,
            jobDesc=jobDesc,
            log=self.log,
            logLevel=logLevel,
        )
        submissionResult = S_OK("Job submitted")

        if submission["OK"]:
            batchID = submission["Value"]
            self.log.info("Job submitted", "(DIRAC JobID: %s; Batch ID: %s" % (jobID, batchID))
            if "PayloadFailed" in submission:
                submissionResult["PayloadFailed"] = submission["PayloadFailed"]
            time.sleep(self.jobSubmissionDelay)
        else:
            self.log.error("Job submission failed", jobID)
            jobReport.setJobParameter(
                par_name="ErrorMessage", par_value="%s CE Submission Error" % (self.ceName), sendFlag=False
            )
            if "ReschedulePayload" in submission:
                self._rescheduleFailedJob(jobID, submission["Message"], self.stopOnApplicationFailure)
                return S_OK()  # Without this, the job is marked as failed
            else:
                if "Value" in submission:  # yes, it's "correct", S_ERROR with 'Value' key
                    self.log.error(
                        "Error in DIRAC JobWrapper or inner CE execution:",
                        "exit code = %s" % (str(submission["Value"])),
                    )
            self.log.error("CE Error", "%s : %s" % (self.ceName, submission["Message"]))
            submissionResult = submission

        return submissionResult

    #############################################################################
    def _getJDLParameters(self, jdl):
        """Returns a dictionary of JDL parameters."""
        try:
            parameters = {}
            #      print jdl
            if not re.search(r"\[", jdl):
                jdl = "[" + jdl + "]"
            classAdJob = ClassAd(jdl)
            paramsDict = classAdJob.contents
            for param, value in paramsDict.items():
                if value.strip().startswith("{"):
                    self.log.debug("Found list type parameter %s" % (param))
                    rawValues = value.replace("{", "").replace("}", "").replace('"', "").split()
                    valueList = []
                    for val in rawValues:
                        if re.search(",$", val):
                            valueList.append(val[:-1])
                        else:
                            valueList.append(val)
                    parameters[param] = valueList
                else:
                    parameters[param] = value.replace('"', "").replace("{", '"{').replace("}", '}"')
                    self.log.debug("Found standard parameter %s: %s" % (param, parameters[param]))
            return S_OK(parameters)
        except Exception as x:
            self.log.exception(lException=x)
            return S_ERROR("Exception while extracting JDL parameters for job")

    #############################################################################
    def __finish(self, message, stop=True):
        """Force the JobAgent to complete gracefully."""
        if stop:
            self.log.info("JobAgent will stop", 'with message "%s", execution complete.' % message)
            self.am_stopExecution()
            return S_ERROR(message)

        return S_OK(message)

    #############################################################################
    def _rescheduleFailedJob(self, jobID, message, stop=True):
        """
        Set Job Status to "Rescheduled" and issue a reschedule command to the Job Manager
        """

        self.log.warn("Failure ==> rescheduling", "(during %s)" % (message))

        jobReport = JobReport(int(jobID), "JobAgent@%s" % self.siteName)

        # Setting a job parameter does not help since the job will be rescheduled,
        # instead set the status with the cause and then another status showing the
        # reschedule operation.

        jobReport.setJobStatus(status=JobStatus.RESCHEDULED, applicationStatus=message, sendFlag=True)

        self.log.info("Job will be rescheduled")
        result = JobManagerClient().rescheduleJob(jobID)
        if not result["OK"]:
            self.log.error("Failed to reschedule job", result["Message"])
            return self.__finish("Problem Rescheduling Job", stop)

        self.log.info("Job Rescheduled", jobID)
        return self.__finish("Job Rescheduled", stop)

    #############################################################################
    def _sendFailoverRequest(self, request):
        """Send failover reques per Job.
        This request would basically be a DISET request for setting the job status.

        If this fails, it only prints a message.

        :param Request request: Request() object
        :return: None
        """
        if len(request):
            self.log.info("Trying to send the failover request")
            # The request is ready, send it now
            isValid = RequestValidator().validate(request)
            if not isValid["OK"]:
                self.log.error("Failover request is not valid", isValid["Message"])
                self.log.error("Printing out the content of the request")
                reqToJSON = request.toJSON()
                if reqToJSON["OK"]:
                    print(str(reqToJSON["Value"]))
                else:
                    self.log.error("Something went wrong creating the JSON from request", reqToJSON["Message"])
            else:
                # Now trying to send the request
                requestClient = ReqClient()
                result = requestClient.putRequest(request)
                if not result["OK"]:
                    self.log.error("Failed to set failover request", result["Message"])

    def finalize(self):
        """Job Agent finalization method"""

        # wait for all jobs to be completed
        res = self.computingElement.shutdown()
        if not res["OK"]:
            self.log.error("CE could not be properly shut down", res["Message"])
        elif res["Value"]:
            self.log.info("Job submission(s) result", res["Value"])

        gridCE = gConfig.getValue("/LocalSite/GridCE", "")
        queue = gConfig.getValue("/LocalSite/CEQueue", "")
        result = PilotManagerClient().setPilotStatus(
            str(self.pilotReference), PilotStatus.DONE, gridCE, "Report from JobAgent", self.siteName, queue
        )
        if not result["OK"]:
            self.log.warn("Issue setting the pilot status", result["Message"])

        return S_OK()



""" Encapsulate here the logic for matching jobs

    Utilities and classes here are used by MatcherHandler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id"

import time
import re

from DIRAC import gLogger, convertToPy3VersionNumber

from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.Core.Utilities.PrettyPrint import printDict
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client.Limiter import Limiter
from DIRAC.WorkloadManagementSystem.Client import PilotStatus

from DIRAC.WorkloadManagementSystem.DB.TaskQueueDB import TaskQueueDB, singleValueDefFields, multiValueMatchFields
from DIRAC.WorkloadManagementSystem.DB.PilotAgentsDB import PilotAgentsDB
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.DB.JobLoggingDB import JobLoggingDB
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus


class PilotVersionError(Exception):
    pass


class Matcher(object):
    """Logic for matching"""

    def __init__(self, pilotAgentsDB=None, jobDB=None, tqDB=None, jlDB=None, opsHelper=None, pilotRef=None):
        """c'tor"""
        if pilotAgentsDB:
            self.pilotAgentsDB = pilotAgentsDB
        else:
            self.pilotAgentsDB = PilotAgentsDB()
        if jobDB:
            self.jobDB = jobDB
        else:
            self.jobDB = JobDB()
        if tqDB:
            self.tqDB = tqDB
        else:
            self.tqDB = TaskQueueDB()
        if jlDB:
            self.jlDB = jlDB
        else:
            self.jlDB = JobLoggingDB()

        if opsHelper:
            self.opsHelper = opsHelper
        else:
            self.opsHelper = Operations()

        if pilotRef:
            self.log = gLogger.getSubLogger("[%s]Matcher" % pilotRef)
            self.pilotAgentsDB.log = gLogger.getSubLogger("[%s]Matcher" % pilotRef)
            self.jobDB.log = gLogger.getSubLogger("[%s]Matcher" % pilotRef)
            self.tqDB.log = gLogger.getSubLogger("[%s]Matcher" % pilotRef)
            self.jlDB.log = gLogger.getSubLogger("[%s]Matcher" % pilotRef)
        else:
            self.log = gLogger.getSubLogger("Matcher")

        self.limiter = Limiter(jobDB=self.jobDB, opsHelper=self.opsHelper, pilotRef=pilotRef)

        self.siteClient = SiteStatus()

    def selectJob(self, resourceDescription, credDict):
        """Main job selection function to find the highest priority job matching the resource capacity"""

        startTime = time.time()

        resourceDict = self._getResourceDict(resourceDescription, credDict)

        # Make a nice print of the resource matching parameters
        toPrintDict = dict(resourceDict)
        if "MaxRAM" in resourceDescription:
            toPrintDict["MaxRAM"] = resourceDescription["MaxRAM"]
        if "NumberOfProcessors" in resourceDescription:
            toPrintDict["NumberOfProcessors"] = resourceDescription["NumberOfProcessors"]
        toPrintDict["Tag"] = []
        if "Tag" in resourceDict:
            for tag in resourceDict["Tag"]:
                if not tag.endswith("GB") and not tag.endswith("Processors"):
                    toPrintDict["Tag"].append(tag)
        if not toPrintDict["Tag"]:
            toPrintDict.pop("Tag")
        self.log.info("Resource description for matching", printDict(toPrintDict))

        negativeCond = self.limiter.getNegativeCondForSite(resourceDict["Site"], resourceDict.get("GridCE"))
        result = self.tqDB.matchAndGetJob(resourceDict, negativeCond=negativeCond)

        if not result["OK"]:
            raise RuntimeError(result["Message"])
        result = result["Value"]
        if not result["matchFound"]:
            self.log.info("No match found")
            return {}

        jobID = result["jobId"]
        resAtt = self.jobDB.getJobAttributes(jobID, ["OwnerDN", "OwnerGroup", "Status"])
        if not resAtt["OK"]:
            raise RuntimeError("Could not retrieve job attributes")
        if not resAtt["Value"]:
            raise RuntimeError("No attributes returned for job")
        if not resAtt["Value"]["Status"] == "Waiting":
            self.log.error("Job matched by the TQ is not in Waiting state", str(jobID))
            result = self.tqDB.deleteJob(jobID)
            if not result["OK"]:
                raise RuntimeError(result["Message"])
            raise RuntimeError("Job %s is not in Waiting state" % str(jobID))

        self._reportStatus(resourceDict, jobID)

        result = self.jobDB.getJobJDL(jobID)
        if not result["OK"]:
            raise RuntimeError("Failed to get the job JDL")

        resultDict = {}
        resultDict["JDL"] = result["Value"]
        resultDict["JobID"] = jobID

        matchTime = time.time() - startTime
        self.log.verbose("Match time", "[%s]" % str(matchTime))
        gMonitor.addMark("matchTime", matchTime)

        # Get some extra stuff into the response returned
        resOpt = self.jobDB.getJobOptParameters(jobID)
        if resOpt["OK"]:
            for key, value in resOpt["Value"].items():
                resultDict[key] = value
        resAtt = self.jobDB.getJobAttributes(jobID, ["OwnerDN", "OwnerGroup"])
        if not resAtt["OK"]:
            raise RuntimeError("Could not retrieve job attributes")
        if not resAtt["Value"]:
            raise RuntimeError("No attributes returned for job")

        if self.opsHelper.getValue("JobScheduling/CheckMatchingDelay", True):
            self.limiter.updateDelayCounters(resourceDict["Site"], jobID)

        pilotInfoReportedFlag = resourceDict.get("PilotInfoReportedFlag", False)
        if not pilotInfoReportedFlag:
            self._updatePilotInfo(resourceDict)
        self._updatePilotJobMapping(resourceDict, jobID)

        resultDict["DN"] = resAtt["Value"]["OwnerDN"]
        resultDict["Group"] = resAtt["Value"]["OwnerGroup"]
        resultDict["PilotInfoReportedFlag"] = True

        return resultDict

    def _getResourceDict(self, resourceDescription, credDict):
        """from resourceDescription to resourceDict (just various mods)"""
        resourceDict = self._processResourceDescription(resourceDescription)
        resourceDict = self._checkCredentials(resourceDict, credDict)
        self._checkPilotVersion(resourceDict)
        if not self._checkMask(resourceDict):
            # Banned destinations can only take Test jobs
            resourceDict["JobType"] = "Test"

        self.log.verbose("Resource description")
        for key in resourceDict:
            self.log.debug("%s : %s" % (key.rjust(20), resourceDict[key]))

        return resourceDict

    def _processResourceDescription(self, resourceDescription):
        """Check and form the resource description dictionary

        :param resourceDescription: a ceDict coming from a JobAgent,
                                    for example.
        :return: updated dictionary of resource description parameters
        """

        resourceDict = {}
        for name in singleValueDefFields:
            if name in resourceDescription:
                resourceDict[name] = resourceDescription[name]

        for name in multiValueMatchFields:
            if name in resourceDescription:
                resourceDict[name] = resourceDescription[name]

        if resourceDescription.get("Tag"):
            tags = resourceDescription["Tag"]
            resourceDict["Tag"] = (
                tags if isinstance(tags, list) else list({tag.strip("\"'") for tag in tags.strip("[]").split(",")})
            )
            if "RequiredTag" in resourceDescription:
                requiredTagsList = (
                    resourceDescription["RequiredTag"].split(",")
                    if isinstance(resourceDescription["RequiredTag"], str)
                    else resourceDescription["RequiredTag"]
                )
                resourceDict["RequiredTag"] = requiredTagsList

        if "JobID" in resourceDescription:
            resourceDict["JobID"] = resourceDescription["JobID"]

        # Convert MaxRAM and NumberOfProcessors parameters into a list of tags
        maxRAM = resourceDescription.get("MaxRAM")
        if maxRAM:
            try:
                maxRAM = int(maxRAM / 1000)
            except ValueError:
                maxRAM = None
        nProcessors = resourceDescription.get("NumberOfProcessors")
        if nProcessors:
            try:
                nProcessors = int(nProcessors)
            except ValueError:
                nProcessors = None
        for param, key in [(maxRAM, "GB"), (nProcessors, "Processors")]:
            if param and param <= 1024:
                paramList = list(range(2, param + 1))
                paramTags = ["%d%s" % (par, key) for par in paramList]
                if paramTags:
                    resourceDict.setdefault("Tag", []).extend(paramTags)

        # Add 'MultiProcessor' to the list of tags
        if nProcessors and nProcessors > 1:
            resourceDict.setdefault("Tag", []).append("MultiProcessor")

        # Add 'WholeNode' to the list of tags
        if "WholeNode" in resourceDescription:
            resourceDict.setdefault("Tag", []).append("WholeNode")

        if "Tag" in resourceDict:
            resourceDict["Tag"] = list(set(resourceDict["Tag"]))
        if "RequiredTag" in resourceDict:
            resourceDict["RequiredTag"] = list(set(resourceDict["RequiredTag"]))

        for k in (
            "DIRACVersion",
            "ReleaseVersion",
            "ReleaseProject",
            "VirtualOrganization",
            "PilotReference",
            "PilotBenchmark",
            "PilotInfoReportedFlag",
        ):
            if k in resourceDescription:
                resourceDict[k] = resourceDescription[k]

        return resourceDict

    def _reportStatus(self, resourceDict, jobID):
        """Reports the status of the matched job in jobDB and jobLoggingDB

        Do not fail if errors happen here
        """
        attNames = ["Status", "MinorStatus", "ApplicationStatus", "Site"]
        attValues = ["Matched", "Assigned", "Unknown", resourceDict["Site"]]
        result = self.jobDB.setJobAttributes(jobID, attNames, attValues)
        if not result["OK"]:
            self.log.error(
                "Problem reporting job status", "setJobAttributes, jobID = %s: %s" % (jobID, result["Message"])
            )
        else:
            self.log.verbose("Set job attributes for jobID", jobID)

        result = self.jlDB.addLoggingRecord(jobID, status="Matched", minorStatus="Assigned", source="Matcher")
        if not result["OK"]:
            self.log.error(
                "Problem reporting job status", "addLoggingRecord, jobID = %s: %s" % (jobID, result["Message"])
            )
        else:
            self.log.verbose("Added logging record for jobID", jobID)

    def _checkMask(self, resourceDict):
        """Check the mask: are we allowed to run normal jobs?

        FIXME: should we move to site OR SE?
        """
        if "Site" not in resourceDict:
            self.log.error("Missing Site Name in Resource JDL")
            raise RuntimeError("Missing Site Name in Resource JDL")

        # Check if site is allowed
        result = self.siteClient.getUsableSites(resourceDict["Site"])
        if not result["OK"]:
            self.log.error("Internal error", "siteClient.getUsableSites: %s" % result["Message"])
            raise RuntimeError("Internal error")

        if resourceDict["Site"] not in result["Value"]:
            return False

        return True

    def _updatePilotInfo(self, resourceDict):
        """Update pilot information - do not fail if we don't manage to do it"""
        pilotReference = resourceDict.get("PilotReference", "")
        if pilotReference and pilotReference != "Unknown":
            gridCE = resourceDict.get("GridCE", "Unknown")
            site = resourceDict.get("Site", "Unknown")
            benchmark = resourceDict.get("PilotBenchmark", 0.0)
            self.log.verbose(
                "Reporting pilot info",
                "for %s: gridCE=%s, site=%s, benchmark=%f" % (pilotReference, gridCE, site, benchmark),
            )

            result = self.pilotAgentsDB.setPilotStatus(
                pilotReference, status=PilotStatus.RUNNING, gridSite=site, destination=gridCE, benchmark=benchmark
            )
            if not result["OK"]:
                self.log.warn(
                    "Problem updating pilot information",
                    "; setPilotStatus. pilotReference: %s; %s" % (pilotReference, result["Message"]),
                )

    def _updatePilotJobMapping(self, resourceDict, jobID):
        """Update pilot to job mapping information"""
        pilotReference = resourceDict.get("PilotReference", "")
        if pilotReference and pilotReference != "Unknown":
            result = self.pilotAgentsDB.setCurrentJobID(pilotReference, jobID)
            if not result["OK"]:
                self.log.error(
                    "Problem updating pilot information",
                    ";setCurrentJobID. pilotReference: %s; %s" % (pilotReference, result["Message"]),
                )
            result = self.pilotAgentsDB.setJobForPilot(jobID, pilotReference, updateStatus=False)
            if not result["OK"]:
                self.log.error(
                    "Problem updating pilot information",
                    "; setJobForPilot. pilotReference: %s; %s" % (pilotReference, result["Message"]),
                )

    def _checkCredentials(self, resourceDict, credDict):
        """Check if we can get a job given the passed credentials"""
        if Properties.GENERIC_PILOT in credDict["properties"]:
            # You can only match groups in the same VO
            if credDict["group"] == "hosts":
                # for the host case the VirtualOrganization parameter
                # is mandatory in resourceDict
                vo = resourceDict.get("VirtualOrganization", "")
            else:
                vo = Registry.getVOForGroup(credDict["group"])
            if "OwnerGroup" not in resourceDict:
                result = Registry.getGroupsForVO(vo)
                if result["OK"]:
                    resourceDict["OwnerGroup"] = result["Value"]
                else:
                    raise RuntimeError(result["Message"])
        else:
            # If it's a private pilot, the DN has to be the same
            if Properties.PILOT in credDict["properties"]:
                self.log.notice("Setting the resource DN to the credentials DN")
                resourceDict["OwnerDN"] = credDict["DN"]
            # If it's a job sharing. The group has to be the same and just check that the DN (if any)
            # belongs to the same group
            elif Properties.JOB_SHARING in credDict["properties"]:
                resourceDict["OwnerGroup"] = credDict["group"]
                self.log.notice("Setting the resource group to the credentials group")
                if "OwnerDN" in resourceDict and resourceDict["OwnerDN"] != credDict["DN"]:
                    ownerDN = resourceDict["OwnerDN"]
                    result = Registry.getGroupsForDN(resourceDict["OwnerDN"])
                    if not result["OK"]:
                        raise RuntimeError(result["Message"])
                    if credDict["group"] not in result["Value"]:
                        # DN is not in the same group! bad boy.
                        self.log.warn(
                            "You cannot request jobs from this DN, as it does not belong to your group!",
                            "(%s)" % ownerDN,
                        )
                        resourceDict["OwnerDN"] = credDict["DN"]
            # Nothing special, group and DN have to be the same
            else:
                resourceDict["OwnerDN"] = credDict["DN"]
                resourceDict["OwnerGroup"] = credDict["group"]

        return resourceDict

    def _checkPilotVersion(self, resourceDict):
        """Check the pilot DIRAC version"""
        if self.opsHelper.getValue("Pilot/CheckVersion", True):
            if "ReleaseVersion" not in resourceDict:
                if "DIRACVersion" not in resourceDict:
                    raise PilotVersionError("Version check requested and not provided by Pilot")
                else:
                    pilotVersion = resourceDict["DIRACVersion"]
            else:
                pilotVersion = resourceDict["ReleaseVersion"]

            validVersions = [
                convertToPy3VersionNumber(newStyleVersion)
                for newStyleVersion in self.opsHelper.getValue("Pilot/Version", [])
            ]
            if validVersions and convertToPy3VersionNumber(pilotVersion) not in validVersions:
                raise PilotVersionError(
                    "Pilot version does not match the production version: %s not in ( %s )"
                    % (pilotVersion, ",".join(validVersions))
                )
            # Check project if requested
            validProject = self.opsHelper.getValue("Pilot/Project", "")
            if validProject:
                if "ReleaseProject" not in resourceDict:
                    raise PilotVersionError(
                        "Version check requested but expected project %s not received" % validProject
                    )
                if resourceDict["ReleaseProject"] != validProject:
                    raise PilotVersionError(
                        "Version check requested but expected project %s != received %s"
                        % (validProject, resourceDict["ReleaseProject"])
                    )



""" ComponentMonitoring class is a front-end to the Component monitoring Database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import random
from six.moves.urllib import parse as urlparse

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import Time, List, Network
from DIRAC.ConfigurationSystem.Client.PathFinder import getSystemURLs

__RCSID__ = "$Id$"


class ComponentMonitoringDB(DB):
    def __init__(self):
        """c'tor

        Initialize the DB
        """
        DB.__init__(self, "ComponentMonitoringDB", "Framework/ComponentMonitoringDB")
        retVal = self.__initializeDB()
        if not retVal["OK"]:
            raise Exception("Can't create tables: %s" % retVal["Message"])
        self.__optionalFields = ("startTime", "cycles", "version", "queries", "DIRACVersion", "description", "platform")
        self.__mainFields = (
            "Id",
            "Setup",
            "Type",
            "ComponentName",
            "Host",
            "Port",
            "StartTime",
            "LastHeartbeat",
            "cycles",
            "queries",
            "LoggingState",
        )
        self.__versionFields = ("VersionTimestamp", "Version", "DIRACVersion", "Platform", "Description")

    def getOptionalFields(self):
        return self.__optionalFields

    def __getTableName(self, name):
        return "compmon_%s" % name

    def __initializeDB(self):
        """
        Create the tables
        """
        retVal = self._query("show tables")
        if not retVal["OK"]:
            return retVal

        tablesInDB = [t[0] for t in retVal["Value"]]
        tablesD = {}

        tN = self.__getTableName("Components")
        if tN not in tablesInDB:
            tablesD[tN] = {
                "Fields": {
                    "Id": "INTEGER AUTO_INCREMENT NOT NULL",
                    "ComponentName": "VARCHAR(255) NOT NULL",
                    "Setup": "VARCHAR(255) NOT NULL",
                    "Type": 'ENUM ( "service", "agent" ) NOT NULL',
                    "Host": "VARCHAR(255) NOT NULL",
                    "Port": "INTEGER DEFAULT 0",
                    "LastHeartbeat": "DATETIME NOT NULL",
                    "StartTime": "DATETIME NOT NULL",
                    "LoggingState": 'VARCHAR(64) DEFAULT "unknown"',
                    "Cycles": "INTEGER",
                    "Queries": "INTEGER",
                },
                "PrimaryKey": "Id",
                "Indexes": {
                    "ComponentIndex": ["ComponentName", "Setup", "Host", "Port"],
                    "TypeIndex": ["Type"],
                },
            }

        tN = self.__getTableName("VersionHistory")
        if tN not in tablesInDB:
            tablesD[tN] = {
                "Fields": {
                    "CompId": "INTEGER NOT NULL",
                    "VersionTimestamp": "DATETIME NOT NULL",
                    "Version": "VARCHAR(255)",
                    "DIRACVersion": "VARCHAR(255) NOT NULL",
                    "Platform": "VARCHAR(255) NOT NULL",
                    "Description": "BLOB",
                },
                "Indexes": {"Component": ["CompId"]},
            }

        return self._createTables(tablesD)

    def __datetime2str(self, dt):
        """
        This method converts the datetime type to a string type.
        """
        if isinstance(dt, six.string_types):
            return dt
        return "%s-%s-%s %s:%s:%s" % (dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second)

    def __registerIfNotThere(self, compDict):
        """
        Registers the component if it's not there
        """
        sqlCond = []
        sqlInsertFields = []
        sqlInsertValues = []
        tableName = self.__getTableName("Components")
        for field in ("componentName", "setup", "type", "host", "port"):
            if field not in compDict:
                if field == "port":
                    continue
                return S_ERROR("Missing %s field in the component dict" % field)
            value = compDict[field]
            field = field.capitalize()
            sqlInsertFields.append(field)
            sqlInsertValues.append("'%s'" % value)
            sqlCond.append("%s = '%s'" % (field, value))
        compLogName = ":".join(sqlInsertValues).replace("'", "")
        self.log.info("Trying to register %s" % compLogName)
        result = self._query("SELECT id FROM `%s` WHERE %s" % (tableName, " AND ".join(sqlCond)))
        if not result["OK"]:
            self.log.error("Cannot register component", "%s: %s" % (compLogName, result["Message"]))
            return result
        if len(result["Value"]):
            compId = result["Value"][0][0]
            self.log.info("%s has compId %s" % (compLogName, compId))
            return S_OK(compId)
        # It's not there, we just need to insert it
        sqlInsertFields.append("LastHeartbeat")
        sqlInsertValues.append("UTC_TIMESTAMP()")
        if "startTime" in compDict:
            sqlInsertFields.append("StartTime")
            val = compDict["startTime"]
            if isinstance(val, Time._allDateTypes):
                val = self.__datetime2str(val)
            sqlInsertValues.append("'%s'" % val)
        for field in ("cycles", "queries"):
            if field not in compDict:
                compDict[field] = 0
            value = compDict[field]
            field = field.capitalize()
            sqlInsertFields.append(field)
            sqlInsertValues.append(str(value))
        self.log.info("Registering component %s" % compLogName)
        result = self._update(
            "INSERT INTO `%s` ( %s ) VALUES ( %s )"
            % (tableName, ", ".join(sqlInsertFields), ", ".join(sqlInsertValues))
        )
        if not result["OK"]:
            return result
        compId = result["lastRowId"]
        self.log.info("%s has compId %s" % (compLogName, compId))
        return S_OK(compId)

    def __updateVersionHistoryIfNeeded(self, compId, compDict):
        """
        Updates the version history given the condition dictionary and component id.
        """
        sqlCond = ["CompId=%s" % compId]
        sqlInsertFields = []
        sqlInsertValues = []
        tableName = self.__getTableName("VersionHistory")
        for field in ("version", "DIRACVersion", "platform"):
            if field not in compDict:
                return S_ERROR("Missing %s field in the component dict" % field)
            value = compDict[field]
            field = field.capitalize()
            sqlInsertFields.append(field)
            sqlInsertValues.append("'%s'" % value)
            sqlCond.append("%s = '%s'" % (field, value))
        result = self._query("SELECT CompId FROM `%s` WHERE %s" % (tableName, " AND ".join(sqlCond)))
        if not result["OK"]:
            return result
        if len(result["Value"]):
            return S_OK(compId)
        # It's not there, we just need to insert it
        sqlInsertFields.append("CompId")
        sqlInsertValues.append(str(compId))
        sqlInsertFields.append("VersionTimestamp")
        sqlInsertValues.append("UTC_TIMESTAMP()")
        if "description" in compDict:
            sqlInsertFields.append("Description")
            result = self._escapeString(compDict["description"])
            if not result["OK"]:
                return result
            sqlInsertValues.append(result["Value"])
        result = self._update(
            "INSERT INTO `%s` ( %s ) VALUES ( %s )"
            % (tableName, ", ".join(sqlInsertFields), ", ".join(sqlInsertValues))
        )
        if not result["OK"]:
            return result
        return S_OK(compId)

    def registerComponent(self, compDict, shallow=False):
        """
        Register a new component in the DB given a component dictionary and returns a component id.
        And if it's already registered it returns the corresponding component id.
        """
        result = self.__registerIfNotThere(compDict)
        if not result["OK"]:
            return result
        compId = result["Value"]
        if shallow:
            return S_OK(compId)
        # Check if something has changed in the version history
        result = self.__updateVersionHistoryIfNeeded(compId, compDict)
        if not result["OK"]:
            return result
        return S_OK(compId)

    def heartbeat(self, compDict):
        """
        Updates the heartbeat
        """
        if "compId" not in compDict:
            result = self.__registerIfNotThere(compDict)
            if not result["OK"]:
                return result
            compId = result["Value"]
            compDict["compId"] = compId
        sqlUpdateFields = ["LastHeartbeat=UTC_TIMESTAMP()"]
        for field in ("cycles", "queries"):
            value = 0
            if field in compDict:
                value = compDict[field]
            sqlUpdateFields.append("%s=%s" % (field.capitalize(), value))
        if "startTime" in compDict:
            sqlUpdateFields.append("StartTime='%s'" % self.__datetime2str(compDict["startTime"]))
        return self._update(
            "UPDATE `%s` SET %s WHERE Id=%s"
            % (self.__getTableName("Components"), ", ".join(sqlUpdateFields), compDict["compId"])
        )

    def __getComponents(self, condDict):
        """
        Loads the components from the DB.

        :type sourceDict: dictionary
        :param sourceDict: The dictionary containing source information.
        :return: S_OK with the components / the error message.
        """
        compTable = self.__getTableName("Components")
        mainFields = ", ".join(self.__mainFields)
        versionTable = self.__getTableName("VersionHistory")
        versionFields = ", ".join(self.__versionFields)
        sqlWhere = []
        for field in condDict:
            val = condDict[field]
            if isinstance(val, six.string_types):
                sqlWhere.append("%s='%s'" % (field, val))
            elif isinstance(val, six.integer_types + (float,)):
                sqlWhere.append("%s='%s'" % (field, val))
            else:
                sqlWhere.append("( %s )" % " OR ".join(["%s='%s'" % (field, v) for v in val]))
        if sqlWhere:
            sqlWhere = "WHERE %s" % " AND ".join(sqlWhere)
        else:
            sqlWhere = ""
        result = self._query("SELECT %s FROM `%s` %s" % (mainFields, compTable, sqlWhere))
        if not result["OK"]:
            return result
        records = []
        dbData = result["Value"]
        for record in dbData:
            rD = {}
            for i in range(len(self.__mainFields)):
                rD[self.__mainFields[i]] = record[i]
            result = self._query(
                "SELECT %s FROM `%s` WHERE CompId=%s ORDER BY VersionTimestamp DESC LIMIT 1"
                % (versionFields, versionTable, rD["Id"])
            )
            if not result["OK"]:
                return result
            if len(result["Value"]) > 0:
                versionRec = result["Value"][0]
                for i in range(len(self.__versionFields)):
                    rD[self.__versionFields[i]] = versionRec[i]
            del rD["Id"]
            records.append(rD)
        return S_OK(StatusSet(records))

    def __checkCondition(self, condDict, field, value):
        """
        It is used to check if a field is present in the condition dictionary or not with the corresponding value.

        :type condDict: dictionary
        :param condDict: The dictionary containing the conditions.
        :type field: string
        :param field: The field.
        :type value: string
        :param field: The value.
        :return: True / False
        """
        if field not in condDict:
            return True
        condVal = condDict[field]
        if isinstance(condVal, (list, tuple)):
            return value in condVal
        return value == condVal

    def __getComponentDefinitionFromCS(self, system, setup, instance, cType, component):
        """
        Gets the basic component details from the configuration file.

        :type system: string
        :param system: The system name.
        :type setup: string
        :param setup: The setup site.
        :type instance: string
        :param instance: The instance.
        :type cType: string
        :param cType: The component type.
        :type component: string
        :param component: The component name.
        :return: a component dictionary.
        """
        componentName = "%s/%s" % (system, component)
        compDict = {"ComponentName": componentName, "Type": cType, "Setup": setup}
        componentSection = "/Systems/%s/%s/%s/%s" % (system, instance, "%ss" % cType.capitalize(), component)
        compStatus = gConfig.getValue("%s/Status" % componentSection, "Active")
        if compStatus.lower() in ("inactive",):
            compDict["Status"] = compStatus.lower().capitalize()
        if cType == "service":
            result = gConfig.getOption("%s/Port" % componentSection)
            if not result["OK"]:
                compDict["Status"] = "Error"
                compDict["Message"] = "Component seems to be defined wrong in CS: %s" % result["Message"]
                return compDict
            try:
                compDict["Port"] = int(result["Value"])
            except Exception:
                compDict["Status"] = "Error"
                compDict["Message"] = "Port for component doesn't seem to be a number"
                return compDict
        return compDict

    def __componentMatchesCondition(self, compDict, requiredComponents, conditionDict={}):
        """
        This method uses __checkCondition method to check if the (key, field) inside component dictionary
        are already present in condition dictionary or not.
        """
        for key in compDict:
            if not self.__checkCondition(conditionDict, key, compDict[key]):
                return False
        return True

    def getComponentsStatus(self, conditionDict={}):
        """
        Get the status of the defined components in the CS compared to the ones that are known in the DB

        :type condDict: dictionary
        :param condDict: The dictionary containing the conditions.
        :return: S_OK with the requires results.
        """
        result = self.__getComponents(conditionDict)
        if not result["OK"]:
            return result
        statusSet = result["Value"]
        requiredComponents = {}
        result = gConfig.getSections("/DIRAC/Setups")
        if not result["OK"]:
            return result
        for setup in result["Value"]:
            if not self.__checkCondition(conditionDict, "Setup", setup):
                continue
            # Iterate through systems
            result = gConfig.getOptionsDict("/DIRAC/Setups/%s" % setup)
            if not result["OK"]:
                return result
            systems = result["Value"]
            for system in systems:
                instance = systems[system]
                # Check defined agents and services
                for cType in ("agent", "service"):
                    # Get entries for the instance of a system
                    result = gConfig.getSections("/Systems/%s/%s/%s" % (system, instance, "%ss" % cType.capitalize()))
                    if not result["OK"]:
                        continue
                    components = result["Value"]
                    for component in components:
                        compDict = self.__getComponentDefinitionFromCS(system, setup, instance, cType, component)
                        if self.__componentMatchesCondition(compDict, requiredComponents, conditionDict):
                            statusSet.addUniqueToSet(requiredComponents, compDict)
                # Walk the URLs
                systemURLs = getSystemURLs(system, setup)  # verify URLs in getSystemURLs method
                for service in systemURLs:
                    # systemURLs is a dict that contain a list of URLs for service
                    if not systemURLs[service]:
                        self.log.error("Not found URL for %s service." % service)
                        continue
                    url = urlparse.urlparse(systemURLs[service][0])
                    if self.__componentMatchesCondition(
                        dict(
                            Setup=setup,
                            Port=url.port,
                            Host=url.hostname,
                            Type="service",
                            ComponentName=system + "/" + service,
                        ),
                        requiredComponents,
                        conditionDict,
                    ):
                        statusSet.addUniqueToSet(requiredComponents, compDict)
        # WALK THE DICT
        statusSet.setComponentsAsRequired(requiredComponents)
        return S_OK(
            (statusSet.getRequiredComponents(), self.__mainFields[1:] + self.__versionFields + ("Status", "Message"))
        )


class StatusSet(object):
    """
    This class is used to set component status as required and this method is used only by the
    ComponentMonitoringDB class.
    """

    def __init__(self, dbRecordsList=[]):
        self.__requiredSet = {}
        self.__requiredFields = ("Setup", "Type", "ComponentName")
        self.__maxSecsSinceHeartbeat = 600
        self.setDBRecords(dbRecordsList)

    def setDBRecords(self, recordsList):
        """
        This sets the DB records given a records list.

        :type recordsList: list
        :param recordsList: a set of records.
        :return: S_OK
        """
        self.__dbSet = {}
        for record in recordsList:
            cD = self.walkSet(self.__dbSet, record)
            cD.append(record)
        return S_OK()

    def addUniqueToSet(self, setDict, compDict):
        """
        Adds unique components to a separate set.

        :type setDict: dictionary
        :param setDict: The set dictionary.
        :type compDict: dictionary
        :param compDict: The dictionary containing the component information.
        """
        rC = self.walkSet(setDict, compDict)
        if compDict not in rC:
            rC.append(compDict)
            inactive = False
            for cD in rC:
                if "Status" in cD and cD["Status"] == "Inactive":
                    inactive = True
                    break
            if inactive:
                for cD in rC:
                    cD["Status"] = "Inactive"

    def walkSet(self, setDict, compDict, createMissing=True):
        """
        Updates the set dictionary.

        :type setDict: dictionary
        :param setDict: The set dictionary.
        :type compDict: dictionary
        :param compDict: The dictionary containing the component information.
        :type creatMissing: bool
        :param createMissing: A variable for adding missing values.
        :return: The set dictionary.
        """
        sD = setDict
        for field in self.__requiredFields:
            val = compDict[field]
            if val not in sD:
                if not createMissing:
                    return None
                if field == self.__requiredFields[-1]:
                    sD[val] = []
                else:
                    sD[val] = {}
            sD = sD[val]
        return sD

    def __reduceComponentList(self, componentList):
        """
        Only keep the most restrictive components.

        :type componentList: list
        :param componentList: A list of components.
        :return: A list of reduced components.
        """
        for i in range(len(componentList)):
            component = componentList[i]
            for j in range(len(componentList)):
                if i == j or componentList[j] is False:
                    continue
                potentiallyMoreRestrictiveComponent = componentList[j]
                match = True
                for key in component:
                    if key not in potentiallyMoreRestrictiveComponent:
                        match = False
                        break
                    if key == "Host":
                        result = Network.checkHostsMatch(component[key], potentiallyMoreRestrictiveComponent[key])
                        if not result["OK"] or not result["Value"]:
                            match = False
                            break
                    else:
                        if component[key] != potentiallyMoreRestrictiveComponent[key]:
                            match = False
                            break
                if match:
                    componentList[i] = False
                    break
        return [comp for comp in componentList if comp]

    def setComponentsAsRequired(self, requiredSet):
        """
        Sets component details according to the required set.

        :type requiredSet: dictionary
        :param requiredSet: The required set dictionary.
        """
        for setup in requiredSet:
            for cType in requiredSet[setup]:
                for name in requiredSet[setup][cType]:
                    # Need to narrow down required
                    cDL = requiredSet[setup][cType][name]
                    cDL = self.__reduceComponentList(cDL)
                    self.__setComponentListAsRequired(cDL)

    def __setComponentListAsRequired(self, compDictList):
        dbD = self.walkSet(self.__dbSet, compDictList[0], createMissing=False)
        if not dbD:
            self.__addMissingDefinedComponents(compDictList)
            return S_OK()
        self.__addFoundDefinedComponent(compDictList)
        return S_OK()

    def __addMissingDefinedComponents(self, compDictList):
        cD = self.walkSet(self.__requiredSet, compDictList[0])
        for compDict in compDictList:
            compDict = self.__setStatus(compDict, "Error", "Component is not up or hasn't connected to register yet")
            cD.append(compDict)

    def __setStatus(self, compDict, status, message=False):
        """
        Sets status within the component dict.

        :type compDict: dictionary
        :param compDict: The component dictionary.
        :type status: string
        :param status: the status.
        :type message: bool
        :param message: the message.
        :return: A component dictionary.
        """
        if "Status" in compDict:
            return compDict
        compDict["Status"] = status
        if message:
            compDict["Message"] = message
        return compDict

    def __addFoundDefinedComponent(self, compDictList):
        cD = self.walkSet(self.__requiredSet, compDictList[0])
        dbD = self.walkSet(self.__dbSet, compDictList[0])
        now = Time.dateTime()
        unmatched = compDictList
        for dbComp in dbD:
            if "Status" not in dbComp:
                self.__setStatus(dbComp, "OK")
                if dbComp["Type"] == "service":
                    if "Port" not in dbComp:
                        self.__setStatus(dbComp, "Error", "Port is not defined")
                    elif dbComp["Port"] not in [compDict["Port"] for compDict in compDictList if "Port" in compDict]:
                        self.__setStatus(
                            compDictList[-1],
                            "Error",
                            "Port (%s) is different that specified in the CS" % dbComp["Port"],
                        )
                elapsed = now - dbComp["LastHeartbeat"]
                elapsed = elapsed.days * 86400 + elapsed.seconds
                if elapsed > self.__maxSecsSinceHeartbeat:
                    self.__setStatus(
                        dbComp,
                        "Error",
                        "Last heartbeat was received at %s (%s secs ago)" % (dbComp["LastHeartbeat"], elapsed),
                    )
            cD.append(dbComp)
            # See if we have a perfect match
            newUnmatched = []
            for unmatchedComp in unmatched:
                perfectMatch = True
                for field in unmatchedComp:
                    if field in ("Status", "Message"):
                        continue
                    if field not in dbComp:
                        perfectMatch = False
                        continue
                    if field == "Host":
                        result = Network.checkHostsMatch(unmatchedComp[field], dbComp[field])
                        if not result["OK"] or not result["Value"]:
                            perfectMatch = False
                    else:
                        if unmatchedComp[field] != dbComp[field]:
                            perfectMatch = False
                if not perfectMatch:
                    newUnmatched.append(unmatchedComp)
            unmatched = newUnmatched
        for unmatchedComp in unmatched:
            self.__setStatus(unmatchedComp, "Error", "There is no component up with this properties")
            cD.append(unmatchedComp)

    def getRequiredComponents(self):
        return self.__requiredSet



""" Here, we need some documentation...
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import os
import types
import threading
import time
import six
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton


@six.add_metaclass(DIRACSingleton)
class Devloader(object):
    def __init__(self):
        self.__log = gLogger.getSubLogger("Devloader")
        self.__reloaded = False
        self.__enabled = True
        self.__reloadTask = False
        self.__stuffToClose = []
        self.__watchedFiles = []
        self.__modifyTimes = {}

    def addStuffToClose(self, stuff):
        self.__stuffToClose.append(stuff)

    @property
    def enabled(self):
        return self.__enabled

    def watchFile(self, fp):
        if os.path.isfile(fp):
            self.__watchedFiles.append(fp)
            return True
        return False

    def __restart(self):
        self.__reloaded = True

        for stuff in self.__stuffToClose:
            try:
                self.__log.always("Closing %s" % stuff)
                sys.stdout.flush()
                stuff.close()
            except Exception:
                gLogger.exception("Could not close %s" % stuff)

        python = sys.executable
        os.execl(python, python, *sys.argv)

    def bootstrap(self):
        if not self.__enabled:
            return False
        if self.__reloadTask:
            return True

        self.__reloadTask = threading.Thread(target=self.__reloadOnUpdate)
        self.__reloadTask.setDaemon(1)
        self.__reloadTask.start()

    def __reloadOnUpdate(self):
        while True:
            time.sleep(1)
            if self.__reloaded:
                return
            for modName in sys.modules:
                modObj = sys.modules[modName]
                if not isinstance(modObj, types.ModuleType):
                    continue
                path = getattr(modObj, "__file__", None)
                if not path:
                    continue
                if path.endswith(".pyc") or path.endswith(".pyo"):
                    path = path[:-1]
                self.__checkFile(path)
            for path in self.__watchedFiles:
                self.__checkFile(path)

    def __checkFile(self, path):
        try:
            modified = os.stat(path).st_mtime
        except Exception:
            return
        if path not in self.__modifyTimes:
            self.__modifyTimes[path] = modified
            return
        if self.__modifyTimes[path] != modified:
            self.__log.always("File system changed (%s). Restarting..." % (path))
            self.__restart()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json

import six


class RMSEncoder(json.JSONEncoder):
    """This class is an encoder for the Requests, Operation and Files."""

    def default(self, obj):  # pylint: disable=method-hidden

        if hasattr(obj, "_getJSONData"):
            return obj._getJSONData()
        elif six.PY3 and isinstance(obj, bytes):
            return obj.decode()
        else:
            return json.JSONEncoder.default(self, obj)



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author: lintao

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division


class HelperReadOnly(object):
    def __init__(self, val):
        self.val = val

    def __get__(self, obj, objtype):
        return self.val

    def __set__(self, obj, val):
        raise AttributeError("can't modify attribute")


class AbsFileSystem(object):
    fs_name = HelperReadOnly("AbsFileSystem")
    seq = HelperReadOnly("/")

    def list_dir(self, path):
        raise NotImplementedError

    def is_dir(self, path):
        raise NotImplementedError


import os
import os.path


class UnixLikeFileSystem(AbsFileSystem):
    fs_name = HelperReadOnly("UnixLikeFileSystem")
    seq = HelperReadOnly("/")

    def list_dir(self, path):
        if not self.is_dir(path):
            raise StopIteration
        for entry in os.listdir(path):
            if self.is_dir(os.path.join(path, entry)):
                entry += self.seq
            yield entry

    def is_dir(self, path):
        return os.path.isdir(path)

    pass


class DFCFileSystem(AbsFileSystem):
    fs_name = HelperReadOnly("DFCFileSystem")
    seq = HelperReadOnly("/")

    def __init__(self, fc):
        self.fc = fc

    def list_dir(self, path):
        if path.endswith("/"):
            path = path.replace("//", "/")
            path = os.path.normpath(path)
        if not self.is_dir(path):
            print("It is not Directory")
            raise StopIteration

        result = self.fc.listDirectory(path, False)
        if not result["OK"]:
            print("some errors.")
            raise StopIteration

        content = result["Value"]["Successful"].get(path, False)
        if not content:
            raise StopIteration

        if content["Files"]:
            for fn in content["Files"]:
                yield self.gen_no_prefix_content(fn, path)
        if content["SubDirs"]:
            for dn in content["SubDirs"]:
                yield self.gen_no_prefix_content(dn, path) + "/"

    def gen_no_prefix_content(self, dn, parent_dn):
        subdn = dn
        if dn.startswith(parent_dn):
            # remove the prefix
            subdn = dn[len(parent_dn) :]
            if subdn.startswith("/"):
                subdn = subdn[1:]
        return subdn

    def is_dir(self, path):
        if path.endswith("/"):
            path = path.replace("//", "/")
            path = os.path.normpath(path)
        result = self.fc.isDirectory(path)
        if not result["OK"]:
            return False
        return result["Value"]["Successful"].get(path, False)


if __name__ == "__main__":
    ulfs = UnixLikeFileSystem()
    print("FS", ulfs.fs_name)
    print("SEQ", ulfs.seq)

    print(list(ulfs.list_dir("/")))
    print(list(ulfs.list_dir("/bad")))



########################################################################
# $HeadURL $
# File: PhysicalRemoval.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/04/02 11:56:10
########################################################################
""" :mod: PhysicalRemoval

    =====================

    .. module: PhysicalRemoval

    :synopsis: PhysicalRemoval operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    PhysicalRemoval operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file PhysicalRemoval.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/04/02 11:56:22
# @brief Definition of PhysicalRemoval class.

# # imports
import os

# # from DIRAC
from DIRAC import S_OK
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase
from DIRAC.Resources.Storage.StorageElement import StorageElement

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class PhysicalRemoval(DMSRequestOperationsBase):
    """
    .. class:: PhysicalRemoval

    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param ~DIRAC.RequestManagementSystem.Client.Operation.Operation operation: Operation instance
        :param str csPath: cs config path
        """
        DMSRequestOperationsBase.__init__(self, operation, csPath)

    def __call__(self):
        """perform physical removal operation"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # gMonitor stuff
            gMonitor.registerActivity(
                "PhysicalRemovalAtt",
                "Physical file removals attempted",
                "RequestExecutingAgent",
                "Files/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "PhysicalRemovalOK",
                "Successful file physical removals",
                "RequestExecutingAgent",
                "Files/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "PhysicalRemovalFail",
                "Failed file physical removals",
                "RequestExecutingAgent",
                "Files/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "PhysicalRemovalSize", "Physically removed size", "RequestExecutingAgent", "Bytes", gMonitor.OP_ACUM
            )

        bannedTargets = self.checkSEsRSS(access="RemoveAccess")
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("PhysicalRemovalAtt")
                gMonitor.addMark("PhysicalRemovalFail")
            return bannedTargets

        if bannedTargets["Value"]:
            return S_OK("%s targets are banned for removal" % ",".join(bannedTargets["Value"]))

        # # get waiting files
        waitingFiles = self.getWaitingFilesList()
        # # prepare lfn dict
        toRemoveDict = dict((opFile.LFN, opFile) for opFile in waitingFiles)

        targetSEs = self.operation.targetSEList

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(toRemoveDict)))
            self.rmsMonitoringReporter.commit()
        else:
            gMonitor.addMark("PhysicalRemovalAtt", len(toRemoveDict) * len(targetSEs))

        # # keep errors dict
        removalStatus = dict.fromkeys(toRemoveDict.keys(), None)
        for lfn in removalStatus:
            removalStatus[lfn] = dict.fromkeys(targetSEs, "")

        for targetSE in targetSEs:

            self.log.info("removing files from %s" % targetSE)

            # # 1st - bulk removal
            bulkRemoval = self.bulkRemoval(toRemoveDict, targetSE)
            if not bulkRemoval["OK"]:
                self.log.error("Failed bulk removal", bulkRemoval["Message"])
                self.operation.Error = bulkRemoval["Message"]
                return bulkRemoval

            bulkRemoval = bulkRemoval["Value"]

            for lfn, opFile in toRemoveDict.items():
                removalStatus[lfn][targetSE] = bulkRemoval["Failed"].get(lfn, "")
                opFile.Error = removalStatus[lfn][targetSE]

            # # 2nd - single file removal
            toRetry = dict((lfn, opFile) for lfn, opFile in toRemoveDict.items() if lfn in bulkRemoval["Failed"])
            for lfn, opFile in toRetry.items():
                self.singleRemoval(opFile, targetSE)
                if not opFile.Error:
                    removalStatus[lfn][targetSE] = ""
                else:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                    else:
                        gMonitor.addMark("PhysicalRemovalFail", 1)
                    removalStatus[lfn][targetSE] = opFile.Error

        # # update file status for waiting files
        failed = 0
        for opFile in self.operation:
            if opFile.Status == "Waiting":
                errors = [error for error in removalStatus[opFile.LFN].values() if error.strip()]
                if errors:
                    failed += 1
                    opFile.Error = ",".join(errors)
                    if "Write access not permitted for this credential" in opFile.Error:
                        opFile.Status = "Failed"

                        if self.rmsMonitoring:
                            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                        else:
                            gMonitor.addMark("PhysicalRemovalFail", len(errors))

                    continue

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                else:
                    gMonitor.addMark("PhysicalRemovalOK", len(targetSEs))
                    gMonitor.addMark("PhysicalRemovalSize", opFile.Size * len(targetSEs))
                opFile.Status = "Done"

        if failed:
            self.operation.Error = "failed to remove %s files" % failed

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK()

    def bulkRemoval(self, toRemoveDict, targetSE):
        """bulk removal of lfns from :targetSE:

        :param dict toRemoveDict: { lfn : opFile, ... }
        :param str targetSE: target SE name
        """

        bulkRemoval = StorageElement(targetSE).removeFile(toRemoveDict)
        return bulkRemoval

    def singleRemoval(self, opFile, targetSE):
        """remove single file from :targetSE:"""
        proxyFile = None
        if "Write access not permitted for this credential" in opFile.Error:
            # # not a DataManger? set status to failed and return
            if "DataManager" not in self.shifter:
                opFile.Status = "Failed"
            elif not opFile.LFN:
                opFile.Error = "LFN not set"
                opFile.Status = "Failed"
            else:
                # #  you're a data manager - save current proxy and get a new one for LFN and retry
                saveProxy = os.environ["X509_USER_PROXY"]
                try:
                    proxyFile = self.getProxyForLFN(opFile.LFN)
                    if not proxyFile["OK"]:
                        opFile.Error = proxyFile["Message"]
                    else:
                        proxyFile = proxyFile["Value"]
                        removeFile = StorageElement(targetSE).removeFile(opFile.LFN)
                        if not removeFile["OK"]:
                            opFile.Error = removeFile["Message"]
                        else:
                            removeFile = removeFile["Value"]
                            if opFile.LFN in removeFile["Failed"]:
                                opFile.Error = removeFile["Failed"][opFile.LFN]
                            else:
                                # # reset error - replica has been removed this time
                                opFile.Error = ""
                finally:
                    if proxyFile:
                        os.unlink(proxyFile)
                    # # put back request owner proxy to env
                    os.environ["X509_USER_PROXY"] = saveProxy
        return S_OK(opFile)



""" This is a test of the creation of the pilot wrapper
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, invalid-name, no-self-use

import os
import base64
import bz2

from DIRAC.WorkloadManagementSystem.Utilities.PilotWrapper import pilotWrapperScript


def test_scriptEmpty():
    """test script creation"""
    res = pilotWrapperScript()

    assert 'cmd = "$py dirac-pilot.py "' in res
    assert 'os.environ["someName"]="someValue"' not in res


def test_scriptoptions():
    """test script creation"""

    res = pilotWrapperScript(
        pilotFilesCompressedEncodedDict={
            "dirac_boh.py": "someContentOfDiracInstall",
            "someOther.py": "someOtherContent",
        },
        pilotOptions="-c 123 --foo bar",
    )

    assert "with open('dirac_boh.py', 'wb') as fd:" in res
    assert 'os.environ["someName"]="someValue"' not in res


def test_scriptReal():
    """test script creation"""

    diracPilot = os.path.join(os.getcwd(), "src/DIRAC/Core/Base/API.py")  # just some random file
    with open(diracPilot, "rb") as fd:
        diracPilot = fd.read()
    diracPilotEncoded = base64.b64encode(bz2.compress(diracPilot, 9))

    diracPilotTools = os.path.join(os.getcwd(), "src/DIRAC/Core/Base/API.py")  # just some random file
    with open(diracPilotTools, "rb") as fd:
        diracPilotTools = fd.read()
    diracPilotToolsEncoded = base64.b64encode(bz2.compress(diracPilotTools, 9))

    diracPilotCommands = os.path.join(os.getcwd(), "src/DIRAC/Core/Base/API.py")  # just some random file
    with open(diracPilotCommands, "rb") as fd:
        diracPilotCommands = fd.read()
    diracPilotCommandsEncoded = base64.b64encode(bz2.compress(diracPilotCommands, 9))

    res = pilotWrapperScript(
        pilotFilesCompressedEncodedDict={
            "dirac-pilot.py": diracPilotEncoded,
            "pilotTools.py": diracPilotToolsEncoded,
            "pilotCommands.py": diracPilotCommandsEncoded,
        },
        pilotOptions="-c 123 --foo bar",
    )

    assert "with open('dirac-pilot.py', 'wb') as fd:" in res
    assert 'os.environ["someName"]="someValue"' not in res


def test_scriptWithEnvVars():
    """test script creation"""
    res = pilotWrapperScript(
        pilotFilesCompressedEncodedDict={
            "dirac_boh.py": "someContentOfDiracInstall",
            "someOther.py": "someOtherContent",
        },
        pilotOptions="-c 123 --foo bar",
        envVariables={"someName": "someValue", "someMore": "oneMore"},
    )

    assert 'os.environ["someName"]="someValue"' in res


def test_scriptPilot3():
    """test script creation"""
    res = pilotWrapperScript(
        pilotFilesCompressedEncodedDict={"proxy": "thisIsSomeProxy"},
        pilotOptions="-c 123 --foo bar",
        envVariables={"someName": "someValue", "someMore": "oneMore"},
        location="lhcb-portal.cern.ch",
    )

    assert 'os.environ["someName"]="someValue"' in res
    assert "lhcb-portal.cern.ch" in res



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-get-input
# Author :  Stuart Paterson
########################################################################
"""
Retrieve input sandbox for DIRAC Job

Example:
  $ dirac-wms-job-get-input 13
  Job input sandbox retrieved in InputSandbox13/
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("D:", "Dir=", "Store the output in this directory")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    sws, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    outputDir = None
    for sw, v in sws:
        if sw in ("D", "Dir"):
            outputDir = v

    for job in parseArguments(args):

        result = dirac.getInputSandbox(job, outputDir=outputDir)
        if result["OK"]:
            if os.path.exists("InputSandbox%s" % job):
                print("Job input sandbox retrieved in InputSandbox%s/" % (job))
        else:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Test_RSS_Policy_DTPolicy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id: $"

from mock import MagicMock
import unittest

from DIRAC import gLogger
import DIRAC.ResourceStatusSystem.Policy.DowntimePolicy as moduleTested


################################################################################


class DTPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup"""
        gLogger.setLevel("DEBUG")
        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.DowntimePolicy

        self.DTCommand = MagicMock()

    def tearDown(self):
        """TearDown"""
        del self.testClass
        del self.moduleTested


################################################################################
# Tests


class DTPolicy_Success(DTPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        policy = self.testClass()
        self.assertEqual("DowntimePolicy", policy.__class__.__name__)

    def test_evaluate(self):
        """tests the evaluate method"""

        policy = self.testClass()

        # command failing
        self.DTCommand.doCommand.return_value = {"OK": False, "Message": "Grumpy command"}
        policy.setCommand(self.DTCommand)
        res = policy.evaluate()
        self.assertTrue(res["OK"])
        self.assertEqual("Grumpy command", res["Value"]["Reason"])
        self.assertEqual("Error", res["Value"]["Status"])

        # command failing /2
        self.DTCommand.doCommand.return_value = {
            "OK": True,
            "Value": {"Severity": "XYZ", "EndDate": "Y", "DowntimeID": "123", "Description": "blah"},
        }
        self.assertEqual("Error", res["Value"]["Status"])

        res = policy.evaluate()
        self.assertTrue(res["OK"])
        # command result empty
        self.DTCommand.doCommand.return_value = {"OK": True, "Value": None}
        res = policy.evaluate()
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("No DownTime announced", res["Value"]["Reason"])

        # command result with a DT
        self.DTCommand.doCommand.return_value = {
            "OK": True,
            "Value": {"Severity": "OUTAGE", "EndDate": "Y", "DowntimeID": "123", "Description": "blah"},
        }
        policy.command = self.DTCommand

        res = policy.evaluate()
        self.assertTrue(res["OK"])
        self.assertEqual("Banned", res["Value"]["Status"])
        self.assertEqual("123 blah", res["Value"]["Reason"])

        # command mock
        self.DTCommand.doCommand.return_value = {
            "OK": True,
            "Value": {"Severity": "WARNING", "EndDate": "Y", "DowntimeID": "123", "Description": "blah"},
        }
        policy.command = self.DTCommand

        res = policy.evaluate()
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("123 blah", res["Value"]["Reason"])


################################################################################
################################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(DTPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(DTPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF
